import copy
import datetime
import sys

from geoformat.conf.error_messages import python_inferior_to_3_7_forbidden
from geoformat.conf.fields_variable import (
    geoformat_field_type_to_python_type,
    recast_black_list,
    none_value_pattern
)
from geoformat.conversion.datetime_conversion import (
    date_to_int,
    time_to_int,
    datetime_to_timestamp,
    int_to_date,
    int_to_time,
    timestamp_to_datetime

)
from geoformat.conf.format_data import (
    value_to_iterable_value,
    is_hexadecimal
)


def update_field_index(fields_metadata, field_name, new_index):
    """
    Updates the index position of a specified field in the fields metadata of a Geolayer.

    This function deep copies the input metadata, checks if the specified field exists, and if so, updates its index
    to a new specified value. It handles index reassignment for all fields to maintain consistency.

    Parameters:
    fields_metadata (dict): A dictionary representing the fields metadata of a Geolayer.
    field_name (str): The name of the field whose index is to be updated.
    new_index (int): The new index position for the specified field.

    Returns:
    dict: The updated fields metadata with the changed index position for the specified field.

    Raises:
    Exception: If the new index is out of valid range or if the specified field name does not exist in the metadata.
    """
    output_fields_metadata = copy.deepcopy(fields_metadata)
    # check if field exists in output_fields_metadata
    if field_name in output_fields_metadata:
        # if "index" key is in output_fields_metadata
        if 'index' in output_fields_metadata[field_name]:
            field_name_original_index = output_fields_metadata[field_name]['index']
        # if not we create it
        else:
            for i_field, field_name_in_metadata in enumerate(output_fields_metadata):
                output_fields_metadata[field_name_in_metadata]['index'] = int(i_field)
                if field_name_in_metadata == field_name:
                    field_name_original_index = i_field

        if new_index != field_name_original_index:
            # check if new index is superior or equal than 0
            if new_index >= 0:
                # check if new index is not superior than existing
                if new_index < len(output_fields_metadata):
                    for i_field, (field_name_in_metadata, field_metadata) in enumerate(output_fields_metadata.items()):
                        # get index value for field_name_in_metadata / create it if not exists
                        field_name_in_metadata_idx = field_metadata['index']

                        # update index for field name
                        if field_name_in_metadata == field_name:
                            output_fields_metadata[field_name_in_metadata]['index'] = new_index
                        else:
                            # re index other fields
                            # minus index when field index is between field_name_original_index and new_index
                            if field_name_original_index < field_name_in_metadata_idx <= new_index:
                                output_fields_metadata[field_name_in_metadata]['index'] -= 1
                                field_name_in_metadata_idx -= 1

                            # plus indexes when field index is between new_index and field_name_original_index
                            if new_index <= field_name_in_metadata_idx < field_name_original_index:
                                output_fields_metadata[field_name_in_metadata]['index'] += 1
                else:
                    raise Exception('new index filed name cannot be superior to {nb_fields}'.format(
                        nb_fields=len(output_fields_metadata) - 1))
            else:
                raise Exception('new index for field name {field_name} must be superior or equal than 0'.format(
                    field_name=field_name))
    else:
        raise Exception('field name {field_name} not exists'.format(field_name=field_name))

    return output_fields_metadata


def recast_field_value(
    field_value,
    recast_value_to_python_type,
    resize_value_width,
    resize_value_precision,
    none_value_pattern=none_value_pattern,
    epoch=datetime.datetime(1970,1,1)
):
    """
    Recasts a field value to a specified Python type, with options for resizing and precision adjustments.

    This function supports recasting various data types including strings, numbers, dates, and times. It handles
    hexadecimal conversion, precision and width resizing, and can process values within lists.

    Parameters:
    field_value: The original value of the field.
    recast_value_to_python_type: The Python type to which the field value is to be recast.
    resize_value_width: The new width for the value, applicable for string and float types.
    resize_value_precision: The new precision for the value, applicable for float types.
    none_value_pattern (set, optional): A set of values considered equivalent to None.
    epoch (datetime.date, optional): The epoch date used for date conversions.

    Returns:
    The recast field value, adjusted for type, precision, and width as specified.

    Raises:
    ValueError: If recasting is not possible due to type incompatibility.
    Exception: If the Python version is not compatible with required features.
    """
    if recast_value_to_python_type == datetime.date and isinstance(epoch, datetime.datetime):
        epoch = epoch.date()

    python_type = recast_value_to_python_type
    collection_type = None

    if isinstance(recast_value_to_python_type, tuple):
        python_type, collection_type = recast_value_to_python_type

    # check if result must be return in list or not
    return_to_list = False
    if collection_type is not None:
        # if value is not None
        if not isinstance(field_value, list):
            if field_value not in none_value_pattern:
                return_to_list = True
        else:
            return_to_list = True

    # put value in list
    if isinstance(field_value, list):
        field_value_list = field_value
        if collection_type is None:
            field_value_list = [python_type(field_value)]
    else:
        field_value_list = [field_value]

    for i_value, field_value in enumerate(field_value_list):
        field_value_type = type(field_value)
        # if value is a list we iterate overt it
        if field_value_type == list:
            field_value_list[i_value] = recast_field_value(
                field_value=field_value,
                recast_value_to_python_type=recast_value_to_python_type,
                resize_value_width=resize_value_width,
                resize_value_precision=resize_value_precision,
                none_value_pattern=none_value_pattern
            )
        else:
            # we cannot retype a None value
            if field_value not in none_value_pattern:
                # recast value
                if field_value_type != python_type:
                    # recast bytes
                    if python_type is bytes and field_value_type is str:
                        # if field_value is hexadecimal we convert it to bytes
                        if is_hexadecimal(field_value):
                            field_value = bytes.fromhex(field_value)
                        else:
                            field_value = eval(field_value)
                    elif python_type is str and field_value_type is bytes:
                        # recast bytes to hexadecimal string
                        field_value = field_value.hex()
                    # recast date, time datetime value
                    elif python_type in {float, int} and field_value_type in {datetime.date, datetime.time, datetime.datetime}:
                        if field_value_type == datetime.date:
                            if isinstance(epoch, datetime.datetime):
                                epoch = epoch.date()
                            field_value = python_type(date_to_int(date_value=field_value, epoch=epoch))
                        elif field_value_type == datetime.time:
                            field_value = python_type(time_to_int(time_value=field_value))
                        else:  # datetime
                            field_value = python_type(datetime_to_timestamp(datetime_value=field_value, epoch=epoch))
                    elif python_type in {datetime.date, datetime.time,
                                                         datetime.datetime} and field_value_type is str:
                        if sys.version_info >= (3, 7):
                            field_value = python_type.fromisoformat(field_value)
                        else:
                            raise Exception(python_inferior_to_3_7_forbidden)
                    elif python_type in {datetime.date, datetime.time,
                                                         datetime.datetime} and field_value_type in {int, float}:
                        if python_type is datetime.date:
                            field_value = int_to_date(int_value=field_value, epoch=epoch)
                        elif python_type is datetime.time:
                            field_value = int_to_time(int_value=field_value)
                        else:  # datetime
                            field_value = timestamp_to_datetime(timestamp=field_value, epoch=epoch)
                    elif python_type is datetime.date and field_value_type is datetime.datetime:
                        field_value = field_value.date()
                    elif python_type is datetime.time and field_value_type is datetime.datetime:
                        field_value = field_value.time()
                    elif python_type is datetime.datetime and field_value_type is datetime.date:
                        field_value = datetime.datetime.combine(field_value, datetime.time.min)
                    # recast boolean
                    elif python_type is bool and field_value_type is str and field_value.upper() == 'FALSE':
                        field_value = False
                    # recast str None value
                    elif field_value_type is str and field_value == '':
                        field_value = None
                    else:
                        try:
                            field_value = python_type(field_value)
                        except ValueError:
                            raise ValueError('value : {field_value} is not compatible with '
                                             '{python_type} type'.format(
                                field_value=field_value,
                                python_type=python_type)
                            )

                # change precision
                if python_type == float and resize_value_precision and field_value:
                    field_value = round(field_value, resize_value_precision)
                # change width
                if python_type in {str, float} and resize_value_width:
                    if field_value is not None:
                        if python_type is float:
                            field_value = str(field_value)
                            resize_value_width += 1  # add comma width
                        field_value = field_value[:resize_value_width]
                        if python_type is float:
                            field_value = float(field_value)
            else:
                field_value = None

            # save value
            field_value_list[i_value] = field_value

    if return_to_list:
        output_value = field_value_list
    else:
        output_value = field_value_list[0]

    return output_value


def recast_field(
    geolayer_to_recast,
    field_name_to_recast,
    recast_to_geoformat_type=None,
    rename_to=None,
    resize_width=None,
    resize_precision=None,
    reindex=None,
    none_value_pattern=none_value_pattern,
    epoch=datetime.datetime(1970,1,1)
):
    """
    Recasts a specified field of a Geolayer to a new type, with options for renaming, resizing, and reindexing.

    The function can change the data type of field, resize its width or precision, change its name, or alter its
    index position. It deeply copies the input layer and updates metadata and values accordingly.

    Parameters:
    geolayer_to_recast (dict): The Geolayer to be modified.
    field_name_to_recast (str): The name of the field to recast.
    recast_to_geoformat_type (str, optional): The new geoformat type to recast the field to.
    rename_to (str, optional): The new name for the field, if renaming is desired.
    resize_width (int, optional): The new width for the field, applicable for certain types.
    resize_precision (int, optional): The new precision for the field, applicable for numerical types.
    reindex (int, optional): The new index position for the field.
    none_value_pattern (set, optional): A set of values considered equivalent to None.
    epoch (datetime.date, optional): The epoch date used for date conversions.

    Returns:
    dict: The modified Geolayer with the recast field.

    Raises:
    Exception: If there are incompatibilities or constraints that prevent the recasting.
    """
    # update metadata
    geolayer_to_recast = copy.deepcopy(geolayer_to_recast)
    input_field_metadata = geolayer_to_recast['metadata']['fields'][field_name_to_recast]
    input_field_type = input_field_metadata['type']
    # check if recasting is compatible with actual type
    if recast_to_geoformat_type in recast_black_list[input_field_type]:
        raise Exception("Input type {input_type} cannot recast to {recast_type} type".format(
            input_type=input_field_type,
            recast_type=recast_to_geoformat_type
        ))

    if recast_to_geoformat_type is not None:
        output_field_metadata = {'type': recast_to_geoformat_type}
    else:
        output_field_metadata = {'type': input_field_metadata['type']}

    # width
    if output_field_metadata['type'] in {'Real', 'RealList', 'String', 'StringList'}:
        if resize_width is not None:
            output_field_metadata['width'] = resize_width
        else:
            if "width" in input_field_metadata:
                output_field_metadata['width'] = input_field_metadata['width']
                # if we swap from Real to String  because of the comma width must take +1
                if 'Real' in input_field_metadata['type'] and 'String' in output_field_metadata['type']:
                    output_field_metadata['width'] += 1
                resize_width = output_field_metadata['width']
            else:
                raise Exception("resize_width must be filled for type : {data_type}".format(
                    data_type=output_field_metadata['type']))

    # precision
    if output_field_metadata['type'] in {'Real', 'RealList'}:
        if resize_precision is not None:
            output_field_metadata['precision'] = resize_precision
        else:
            if "precision" in input_field_metadata:
                output_field_metadata['precision'] = input_field_metadata['precision']
                resize_precision = output_field_metadata['precision']
            else:
                raise Exception(
                    "resize_precision must be filled for {data_type} type".format(
                        data_type=output_field_metadata['type']))

    # add index if exists
    if 'index' in input_field_metadata:
        output_field_metadata['index'] = input_field_metadata['index']

    # write output metadata in geolayer_to_recast
    if input_field_metadata != output_field_metadata:
        geolayer_to_recast['metadata']['fields'][field_name_to_recast] = output_field_metadata

    # re index
    if reindex is not None:
        output_fields_metadata = update_field_index(
            fields_metadata=geolayer_to_recast['metadata']['fields'],
            field_name=field_name_to_recast,
            new_index=reindex
        )
        geolayer_to_recast['metadata']['fields'] = output_fields_metadata

    # rename field in metadata
    if rename_to is not None:
        if rename_to in geolayer_to_recast['metadata']['fields']:
            raise Exception('field {rename_to} still exists on geolayer'.format(rename_to=rename_to))
        if rename_to != field_name_to_recast:
            geolayer_to_recast['metadata']['fields'][rename_to] = geolayer_to_recast['metadata']['fields'][
                field_name_to_recast]
            del geolayer_to_recast['metadata']['fields'][field_name_to_recast]

    # define python type to recast
    recast_to_python_type = geoformat_field_type_to_python_type[output_field_metadata['type']]

    # value in list or not
    if output_field_metadata['type'] in {'IntegerList', 'RealList', 'StringList'}:
        in_list = True
    else:
        in_list = False

    # loop over features
    for i_feat, feature in geolayer_to_recast['features'].items():
        # if attributes in feature
        if 'attributes' in feature:
            if field_name_to_recast in feature['attributes']:
                feature_field_name_value = feature['attributes'][field_name_to_recast]
                feature_field_name_type = type(feature_field_name_value)
                if feature_field_name_type == list:
                    # if value in list and output is not in list we have to recast value in str
                    if in_list is False:
                        feature_field_name_value = str(feature_field_name_value)
                elif feature_field_name_type == dict:
                    feature_field_name_value =  str(feature_field_name_value)
                else:
                    if feature_field_name_value in none_value_pattern:
                        feature_field_name_value = None
                    else:
                        # if result must be in list
                        if in_list is True:
                            try:
                                eval_feature_field_name_value = eval(feature_field_name_value)
                                # force eval_feature_field_name_value to be in list type
                                eval_feature_field_name_value = value_to_iterable_value(eval_feature_field_name_value, list)
                            except (TypeError, NameError, ValueError, SyntaxError):
                                eval_feature_field_name_value = [feature_field_name_value]
                            # put value in list in any case
                            if isinstance(eval_feature_field_name_value, list):
                                feature_field_name_value = eval_feature_field_name_value
                            else:
                                feature_field_name_value = [eval_feature_field_name_value]

                # recast value(s)
                feature_field_name_value = recast_field_value(
                    field_value=feature_field_name_value,
                    recast_value_to_python_type=recast_to_python_type,
                    resize_value_width=resize_width,
                    resize_value_precision=resize_precision,
                    none_value_pattern=none_value_pattern,
                    epoch=epoch
                )

                # saving recasting data
                feature['attributes'][field_name_to_recast] = feature_field_name_value

                # update field_name change
                if rename_to is not None:
                    feature['attributes'][rename_to] = feature['attributes'][field_name_to_recast]
                    del feature['attributes'][field_name_to_recast]

    return geolayer_to_recast



import string


def value_to_iterable_value(value, output_iterable_type=list):
    """
    Transform value to iterate value (list by default)

    :param value: value that we want to transfomr
    :param output_iterable_type: iterable type to transform
    :return: input value transform to given (by iterable_type) iterable value type
    """

    # check if iterable type if iterable
    if output_iterable_type not in {list, tuple, set}:
        raise Exception("output_iterable_type must be list, tuple or set")

    # make type transform
    if value is not None:
        # if output type is same that input
        if isinstance(value, output_iterable_type):
            output_value = value
        else:
            # if input type is yet iterable
            if isinstance(value, (list, tuple, set)):
                value = list(value)
            else:
                # if not make it iterable
                value = [value]
            # transform input type
            output_value = output_iterable_type(value)
    else:
        output_value = None

    return output_value


def is_hexadecimal(s):
    """
    Check if input string value is hexadecimal or not

    :param s: input string
    :return: True if hexadecimal / False if not hexadecimal
    """
    hex_digits = set(string.hexdigits)
    # if s is long, then it is faster to check against a set
    return all(c in hex_digits for c in s)


def pairwise(iterable):
    """
    from https://stackoverflow.com/questions/5764782/iterate-through-pairs-of-items-in-a-python-list?lq=1
    iterable = [s0, s1, s2, s3, ...]
    return ((s0,s1), (s1,s2), (s2, s3), ...)

    """
    import itertools

    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)



import datetime


def format_datetime_object_to_str_value(
    datetime_value, format=["year", "month", "day"]
):
    """
    Convert a datetime object to a string representation based on specified components.

    This function allows selective formatting of a datetime object into a string by choosing specific components
    such as year, month, day, hour, second, and microsecond. The components are concatenated in the order
    provided in the format list.

    Parameters:
    datetime_value (datetime): The datetime object to be formatted.
    format (list of str, optional): A list of string components to include in the output.
        Default is ["year", "month", "day"]. Valid components are "year", "month", "day", "hour",
        "second", and "microsecond".

    Returns:
    str: A string representation of the datetime object based on the specified format.

    Raises:
    Exception: If an invalid component is specified in the format list.
    """

    if isinstance(format, str):
        format = [format]

    return_value = ""
    for v in format:
        if v == "year":
            return_value += str(datetime_value.year).zfill(4)
        elif v == "month":
            return_value += str(datetime_value.month).zfill(2)
        elif v == "day":
            return_value += str(datetime_value.day).zfill(2)
        elif v == "hour":
            return_value += str(datetime_value.hour).zfill(2)
        elif v == "second":
            return_value += str(datetime_value.second).zfill(2)
        elif v == "microsecond":
            return_value += str(datetime_value.microsecond).zfill(6)
        else:
            raise Exception()

    return return_value


def date_to_int(date_value, epoch=datetime.date(1970, 1, 1)):
    """
    Convert a date object to an integer representing the number of days since a specified epoch.

    This function calculates the number of days from a given epoch (default is 1970-01-01) to the specified date.
    It's useful for converting date values into a consistent numeric format for calculations or storage.

    Parameters:
    date_value (datetime.date): The date value to be converted.
    epoch (datetime.date, optional): The epoch (start date) from which the days are counted.
        Default is 1970-01-01.

    Returns:
    int: The number of days from the epoch to the given date.
    """
    return (date_value - epoch).days


def time_to_int(time_value):
    """
    Convert a time object to an integer representing the time in microseconds.

    This function transforms a time object into an integer representing the total number of microseconds
    since midnight. It's useful for time calculations or storing time values in a compact numeric format.

    Parameters:
    time_value (datetime.time): The time value to be converted.

    Returns:
    int: The number of microseconds since midnight represented by the time object.
    """
    return int(
        time_value.hour * 3600 * 1e6
        + time_value.minute * 60 * 1e6
        + time_value.second * 1e6
        + time_value.microsecond
    )


def int_to_date(int_value, epoch=datetime.date(1970, 1, 1)):
    """
    Convert an integer representing the number of days since a specified epoch to a date object.

    Parameters:
    int_value (int): The number of days since the epoch.
    epoch (datetime.date, optional): The epoch (start date) from which the days are counted.
        Default is 1970-01-01.

    Returns:
    datetime.date: The date corresponding to the given number of days since the epoch.
    """
    return epoch + datetime.timedelta(days=int_value)


def int_to_time(int_value):
    """
    Convert an integer representing time in microseconds since midnight to a time object.

    Parameters:
    int_value (int): The number of microseconds since midnight.

    Returns:
    datetime.time: The time object represented by the given number of microseconds since midnight.
    """
    seconds, int_value = divmod(int_value, 1e6)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    return datetime.time(int(hours), int(minutes), int(seconds), int(int_value))


def datetime_to_timestamp(
    datetime_value, epoch=datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
):
    """
    Converts a datetime object to a UNIX timestamp (float) representing seconds since a specified epoch.

    This function takes a datetime object and calculates the number of seconds elapsed from the specified epoch
    to the datetime value. It provides support for both timezone-aware and naive (without timezone information)
    datetime objects. When handling naive datetime objects, they are assumed to be in the local timezone.
    Additionally, the function aligns the timezone of the epoch with that of the datetime_value before performing
    the calculation to ensure accuracy.

    Parameters:
    - datetime_value (datetime.datetime): The datetime value to be converted to a timestamp.
    - epoch (datetime.datetime, optional): The reference epoch time. Defaults to January 1, 1970, 00:00:00 UTC.

    Returns:
    - float: The UNIX timestamp representing the number of seconds from the epoch to the given datetime.

    Notes:
    - This function handles timezone alignment by standardizing the timezones of both the datetime_value and the epoch.
    - Naive datetime objects are treated as being in the local timezone for conversion.
    - Timezone-aware datetime objects are used as is, with the epoch being converted to their timezone for accurate calculation.

    Internal Functions:
    - format_timezones: Aligns the timezones of datetime_value and epoch for consistent calculation.
    - format_timezone: Sets or replaces the timezone information for a given datetime object.
    """

    def format_timezones(datetime_value, epoch_value):
        """
        Aligns the timezones of two datetime objects, either by setting a naive datetime object to a specific timezone
        or ensuring both datetime objects share the same timezone.

        This function checks the timezone information of both the datetime_value and the epoch_value. If one of them is
        timezone-aware and the other is naive (without timezone information), it assigns the timezone of the aware object
        to the naive one. If both are naive, no timezone is assigned. The function ensures that both datetime objects
        are either timezone-aware and aligned, or both remain naive.

        Parameters:
        - datetime_value (datetime.datetime): The primary datetime object to be aligned with the epoch.
        - epoch_value (datetime.datetime): The reference epoch datetime object for alignment.

        Returns:
        - tuple: A tuple containing two datetime.datetime objects with aligned timezones.

        Note:
        - If both datetime_value and epoch_value are timezone-aware but with different timezones, no conversion is done.
        - This function is typically used in timestamp calculation to ensure consistency in timezone handling.
        """
        datetime_value_timezone = datetime_value.tzinfo
        epoch_timezone = epoch_value.tzinfo
        if datetime_value_timezone is not None and epoch_timezone is None:
            timezone = datetime_value_timezone
        elif datetime_value_timezone is None and epoch_timezone is not None:
            timezone = epoch_timezone
        else:
            timezone = None

        datetime_value = format_timezone(
            datetime_value=datetime_value, timezone=timezone
        )
        epoch_value = format_timezone(datetime_value=epoch_value, timezone=timezone)

        return datetime_value, epoch_value

    def format_timezone(datetime_value, timezone):
        """
        Sets or replaces the timezone information of a datetime object.

        This function takes a datetime object and a timezone object. If the datetime object is naive (without timezone
        information), it assigns the provided timezone to it. If the datetime object is already timezone-aware, it
        replaces its existing timezone with the provided one. This function is used to standardize the timezone of
        a datetime object for consistent time-based calculations.

        Parameters:
        - datetime_value (datetime.datetime): The datetime object to which the timezone information will be applied.
        - timezone (datetime.timezone or None): The timezone to be set for the datetime object. If None, the function
          leaves the datetime object unchanged.

        Returns:
        - datetime.datetime: The datetime object with updated timezone information.

        Note:
        - This function does not perform any timezone conversion. It only sets or replaces the timezone information.
        """
        if timezone is not None:
            datetime_value = datetime_value.replace(tzinfo=timezone)

        return datetime_value

    datetime_value, epoch = format_timezones(
        datetime_value=datetime_value, epoch_value=epoch
    )

    # Calculate the difference in seconds
    delta = datetime_value - epoch
    timestamp = delta.total_seconds()

    return timestamp


def _determine_timestamp_unit(timestamp):
    """
    Determine if the given timestamp is in seconds or milliseconds.

    Args:
        timestamp (int or float): The timestamp to evaluate.

    Returns:
        str: 'seconds' if the timestamp is determined to be in seconds,
             'milliseconds' if the timestamp is determined to be in milliseconds.
    """
    # Define a threshold year, beyond which we consider the timestamp to be in milliseconds.
    # This is based on a rough estimate; timestamps representing dates beyond this year in seconds
    # are unlikely, thus considered to be in milliseconds.
    threshold_year = 3000

    try:
        # Try to convert the timestamp to datetime assuming it's in seconds.
        date_in_seconds = datetime.datetime.fromtimestamp(timestamp)
        if date_in_seconds.year > threshold_year:
            # If the year is beyond the threshold, it's likely the input was in milliseconds.
            return "milliseconds"
        else:
            return "seconds"
    except ValueError:
        # If conversion assuming seconds fails (e.g., value is too large), assume milliseconds.
        # This block also handles cases where the timestamp is negative and far in the past.
        return "milliseconds"
    except OverflowError:
        # If the timestamp value is too large even for milliseconds representation,
        # this block catches the overflow error.
        return "milliseconds"


def timestamp_to_datetime(timestamp, epoch=datetime.datetime(1970, 1, 1)):
    """
    Converts a numeric timestamp to a datetime object, automatically determining
    whether the timestamp is in seconds or milliseconds, and adjusting the conversion accordingly.

    This function leverages the _determine_timestamp_unit helper to decide if the given
    timestamp is measured in seconds or milliseconds. It then calculates the datetime
    corresponding to the timestamp, taking into account the unit of measurement and the
    specified epoch. The conversion accounts for timezone information of the epoch.

    Args:
        timestamp (int or float): The timestamp to be converted, which can be in seconds or milliseconds.
        epoch (datetime.datetime, optional): The reference epoch from which time is counted.
                                             Defaults to January 1, 1970, 00:00:00 UTC.

    Returns:
        datetime.datetime: The datetime object corresponding to the given timestamp, in the same timezone as the epoch.

    Notes:
        - The function automatically adjusts for the unit of the input timestamp (seconds or milliseconds).
        - The resulting datetime object will be in the same timezone as the provided epoch.
    """
    # Determine if timestamp is in seconds or milliseconds
    unit = _determine_timestamp_unit(timestamp)
    if unit == "milliseconds":
        # Convert milliseconds to seconds
        timestamp = timestamp / 1000.0

    # Ensure the epoch has timezone information
    if epoch.tzinfo is None:
        epoch = epoch.replace(tzinfo=datetime.timezone.utc)

    # Calculate the datetime corresponding to the timestamp
    return epoch + datetime.timedelta(seconds=timestamp)



import copy
from geoformat.conf.error_messages import (
    field_exists
)
from geoformat.conf.format_data import value_to_iterable_value

def rename_field_in_feature(feature, old_field_name, new_field_name):
    """
    Rename field in feature

    :param feature: input feature.
    :param old_field_name: actual name of field that we want to change.
    :param new_field_name: new field's name that we want to apply.
    :return: feature with field rename
    """
    feature = copy.deepcopy(feature)
    if feature_attributes := feature.get("attributes"):
        if old_field_name in feature_attributes:
            if new_field_name not in feature_attributes:
                feature_attributes[new_field_name] = feature_attributes[old_field_name]
                del feature_attributes[old_field_name]
            else:
                raise Exception(field_exists.format(field_name=new_field_name))

    return feature

def drop_field_in_feature(feature, field_name_or_field_name_list):
    """
    Drop field (if exists) in feature

    :param feature: feature on which we want to delete field(s).
    :param field_name_or_field_name_list: name or name list of field that we want to delete.
    :return: feature without field that we want to delete.
    """
    field_name_or_field_name_list = value_to_iterable_value(field_name_or_field_name_list, output_iterable_type=list)

    feature = copy.deepcopy(feature)
    if feature_attributes := feature.get('attributes'):
        for field_name_to_drop in field_name_or_field_name_list:
            if field_name_to_drop in feature_attributes:
                del feature_attributes[field_name_to_drop]

        # delete attributes key if there is no attributes data in feature
        if not feature_attributes:
            del feature['attributes']

    return feature


def drop_field_that_not_exists_in_feature(feature, not_deleting_field_name_or_field_name_list):
    """
    Drop field in feature that are not specified in not_deleting_field_name_or_field_name_list variable.

    :param feature:  feature on which we want to delete field(s).
    :param not_deleting_field_name_or_field_name_list: name or name list of field that we want to keep.
    :return: feature with field that we want to keep.
    """
    not_deleting_field_name_or_field_name_list  = value_to_iterable_value(
        value=not_deleting_field_name_or_field_name_list,
        output_iterable_type=set
    )
    if feature_attributes := feature.get('attributes'):
        feature_field_name_set = set(list(feature_attributes.keys()))
        field_name_to_drop_set = feature_field_name_set - not_deleting_field_name_or_field_name_list
        for field_name_to_drop in field_name_to_drop_set:
            feature = drop_field_in_feature(feature=feature, field_name_or_field_name_list=field_name_to_drop)

    return feature


from geoformat.conversion.geometry_conversion import (
    geometry_to_geometry_collection,
    multi_geometry_to_single_geometry
)
from geoformat.geoprocessing.connectors.operations import (
    coordinates_to_segment
)

from geoformat.geoprocessing.measure.mesure_length import (
    segment_length
)

def geometry_length(geometry, distance_type='EUCLIDEAN'):

    # initialize return
    length = 0

    # transform function to geometry collection
    geometry = geometry_to_geometry_collection(geometry)
    # if geometry is not empty
    if geometry:
        for geometry in geometry['geometries']:
            # if geometry is not a point or multipoint (they are no dimension then no length)
            if 'POINT' not in geometry['type'].upper():
                # loop over part
                for geometry_part in multi_geometry_to_single_geometry(geometry, bbox=False):
                    coordinates_list = geometry_part['coordinates']
                    # if there is coordinates
                    if coordinates_list:
                        # transform coordinates_list to segment and calculate segment lenght
                        for segment in coordinates_to_segment(coordinates_list):
                            # add distance to
                            length += segment_length(segment, distance_type=distance_type)

        return length


import copy
from pathlib import Path

try:
    import shapefile
    import_shapefile_success = True
except ImportError:
    import_shapefile_success = False


from geoformat.conf.fields_variable import (
    geoformat_field_type_to_shapefile_field_type,
    shapefile_field_type_to_geoformat_field_type,
    python_type_to_geoformat_field_type,
    shapefile_str_max_width,
    field_metadata_width_required,
    field_metadata_precision_required,
)
from geoformat.conf.format_data import value_to_iterable_value
from geoformat.conf.path import verify_input_path_is_dir, path_to_file_path
from geoformat.conf.error_messages import (
    import_shapefile_error,
    geometry_type_does_not_allowed_in_geolayer,
    geometry_type_not_allowed,
    key_parameter_invalid,
    field_name_length_not_valid,
)
from geoformat.conf.geometry_variable import GEOFORMAT_GEOMETRY_TYPE
from geoformat.conversion.coordinates_conversion import format_coordinates
from geoformat.conversion.feature_conversion import (
    features_fields_type_scan,
    feature_filter,
)
from geoformat.manipulation.geolayer_manipulation import feature_list_to_geolayer
from geoformat.conversion.fields_conversion import recast_field_value
from geoformat.conversion.geometry_conversion import geometry_to_bbox, force_rhr
from geoformat.conversion.metadata_conversion import (
    from_field_scan_determine_field_width,
    get_field_name_list_ordered_by_i_field,
)
from geoformat.manipulation.geolayer_manipulation import (
    rename_field,
    split_geolayer_by_geometry_type,
)
from geoformat.manipulation.metadata_manipulation import (
    drop_field_that_not_exists_in_metadata,
)

from geoformat.geoprocessing.geoparameters.bbox import bbox_union
from geoformat.conf.fields_variable import none_value_pattern

shapefile_driver_function_dict = {
    "driver_formating": {
        "max_field_name_char": 10,
        "geometry_mapping": {
            "Point": "Point",
            "LineString": "LineString",
            "Polygon": "Polygon",
            "MultiPoint": "MultiPoint",
            "MultiLineString": "LineString",
            "MultiPolygon": "Polygon",
            "GeometryCollection": "SPLIT_BY_GEOMETRY_TYPE",
        },
        "geometry_type": {
            None: 0,
            "Point": 1,
            "LineString": 3,
            "MultiLineString": 3,
            "Polygon": 5,
            "MultiPolygon": 5,
            "MultiPoint": 8,
        },
        "field_recast": True,
        "format_field_type": {
            "IntegerList": str,
            "RealList": str,
            "StringList": str,
            "Time": str,
            "DateTime": str,
            "Binary": str,
        },
        "field_type_necessary_metadata": {
            "C": {"width"},
            "N": {"width", "precision"},
            "F": {"width", "precision"},
            "L": None,
            "D": None,
        },
        "format_feature_data": {
            "D": {
                "module": "geoformat",
                "function_name": "format_datetime_object_to_str_value",
                "need_parameters_name": "datetime_value",
                "default_parameters": {"format": ["year", "month", "day"]},
            }
        },
    },
    "layer_creator": {
        "create_layer": {
            "module": "shapefile",
            "function_name": "Writer",
            "need_parameters": {"target": "PATH", "shapeType": "OUTPUT_GEOMETRY_TYPE"},
            "default_parameters": {"autoBalance": False},
        },
        "create_field": {
            "function_name": "field",
            "module": "create_layer",
            "method_child_from": "create_layer",
            "need_parameters_name": {
                "name": "new_field_name",
                "fieldType": "type",
                "size": "width",
                "decimal": "precision",
            },
            "default_field": {"id": {"type": "Integer"}},
        },
        "create_feature_attributes": {
            "function_name": "record",
            "module": "create_layer",
            "need_parameters_name": "recordList",  # we can use recordDict too
            "method_child_from": "create_layer",
        },
        "create_feature_geometry": {
            "function_name": "shape",
            "need_parameters_name": "recordList",  # we can use recordDict too
            "method_child_from": "create_layer",
        },
    },
}

pyshp_geometry_type = {
    0: None,
    1: "Point",
    3: "LineString",
    5: "Polygon",
    8: "MultiPoint",
}


def _get_module(metadata_dict, locals_var=None):
    """
    Return module wanted in metadata dict.

    :param metadata_dict:  dict with information to catch function
    :param locals_var: local variable
    :return: return wanted module
    """

    if locals_var is None:
        locals_var = {}

    module_name = metadata_dict.get("module")

    if module_name in locals_var:
        module = locals_var[module_name]
    # if module is string we have to import it
    else:
        module = __import__(
            name=module_name,
            locals=None,
            globals=None,
            fromlist=metadata_dict.get("function_name", ()),
            level=0,
        )

    return module


def _get_function(metadata_dict, locals_var=None):
    """
    Return function wanted in metadata_dict.

    :param metadata_dict: dict with information to catch function
    :param locals_var: script local variables and functions (get with locals() function).
    :return: return wanted function
    """
    module = _get_module(metadata_dict, locals_var=locals_var)
    function_name = metadata_dict["function_name"]
    # get function
    function = getattr(module, function_name)

    return function


def _format_coordinates_for_pyshp(geometry):
    """
    return geometry coordinates dict formatted to PyShp lib formatting coordinates.

    :param: Geoformat geometry.
    :return: dict with input variable with associated data in it for writing geometry in pyshp.
    """
    geometry_type = geometry.get("type")
    if geometry_type.upper() in GEOFORMAT_GEOMETRY_TYPE:
        geometry_coordinates = geometry.get("coordinates")
        return_coordinates = {}
        if geometry_type == "Point":
            return_coordinates = {
                "x": geometry_coordinates[0],
                "y": geometry_coordinates[1],
            }
        elif geometry_type == "LineString":
            return_coordinates = {"lines": [geometry_coordinates]}
        elif geometry_type == "Polygon":
            return_coordinates = {"polys": geometry_coordinates}
        elif geometry_type == "MultiPoint":
            return_coordinates = {"points": geometry_coordinates}
        elif geometry_type == "MultiLineString":
            return_coordinates = {"lines": geometry_coordinates}
        elif geometry_type == "MultiPolygon":
            flat_coordinates = []
            for poly in geometry_coordinates:
                flat_coordinates = flat_coordinates + poly
            return_coordinates = {"polys": flat_coordinates}
    else:
        raise Exception(geometry_type_not_allowed.format(geometry_type=geometry_type))

    return return_coordinates


def _esri_shapefile_check_fields_metadata(_geolayer, _driver_parameters):
    """
    Scan geolayer metadata and use driver parameters to check if fields must be recast or not.
    If field(s) must be recast this function return a dict formatted like this :
        {
            field_name: {     # field name that must be recast
                "recast_value_to_python_type": # python type of recasting,
                "recast_value_to_output_type": # output format type (depend on driver)
                "resize_value_width": # new value's width,
                "resize_value_precision": # new value's precision}
        }

    :param _geolayer: input geolayer
    :param _driver_parameters: esri shapefile driver parameters
    :return: dict (describes below) that contains field name to recast and parameters for recasting
    """
    field_recast = _driver_parameters["driver_formating"].get("field_recast", False)

    # scan field metadata to check if field must have to be recast
    fields_metadata = _geolayer["metadata"].get("fields", {})
    recast_field_mapping = {}
    recast_field_scan_list = []
    for field_name, field_metadata in fields_metadata.items():
        # check if field geoformat must be recast in output format
        recast_to_python_type = _driver_parameters["driver_formating"][
            "format_field_type"
        ].get(field_metadata["type"], None)
        if recast_to_python_type:
            recast_field_mapping[field_name] = {
                "recast_value_to_python_type": recast_to_python_type,
                "recast_value_to_output_type": geoformat_field_type_to_shapefile_field_type[
                    python_type_to_geoformat_field_type[recast_to_python_type]
                ],
                "resize_value_width": None,
                "resize_value_precision": None,
            }
        # if recasting is necessary
        if recast_to_python_type is not None:
            if field_recast is True:
                recast_field_scan_list.append(field_name)

    # scan all fields that will be recast to have information about (width and precision)
    if field_recast and recast_field_scan_list:
        # scan values that must be recast
        field_scan = features_fields_type_scan(
            geolayer_or_feature_list=_geolayer,
            field_name_filter=recast_field_scan_list,
            try_to_force_type=False,
        )

        # put result in output dict
        for field_name in recast_field_scan_list:
            field_name_scan_dict = field_scan[field_name]
            recast_field_name_dict = recast_field_mapping[field_name]
            recast_to_python_type = recast_field_name_dict[
                "recast_value_to_python_type"
            ]
            recast_to_shapefile_type = recast_field_name_dict[
                "recast_value_to_output_type"
            ]
            # get needed metadata
            output_field_type_metadata_need = _driver_parameters["driver_formating"][
                "field_type_necessary_metadata"
            ][recast_to_shapefile_type]
            if "width" in output_field_type_metadata_need:
                field_in_list = False
                if isinstance(recast_to_python_type, tuple):
                    field_in_list = True

                field_name_width = from_field_scan_determine_field_width(
                    field_type=recast_to_python_type,
                    field_in_list=field_in_list,
                    field_width_str=field_name_scan_dict["field_width_str"],
                    field_width_float=field_name_scan_dict["field_width_float"],
                    field_width_list=field_name_scan_dict["field_width_list"],
                    max_width=shapefile_str_max_width,
                )
                recast_field_name_dict["resize_value_width"] = field_name_width

            if "precision" in output_field_type_metadata_need:
                recast_field_name_dict["resize_value_precision"] = field_name_scan_dict[
                    "field_precision"
                ]

    return recast_field_mapping


def _geoformat_feature_to_shapefile_feature(
    feature,
    shapefile_fields_metadata,
    recast_field_mapping,
    shapefile_format_value,
    locals_var=None,
):
    """
    Convert geoformat feature to shapefile feature.

    :param feature: geoformat feature
    :param shapefile_fields_metadata: shapefile field object
    :param recast_field_mapping: dict with field to recast and parameters to recast it
    :param shapefile_format_value: dict that describe how to reformat certain type of shapefile value to be compatible
     with geoformat values
    :param locals_var: locals variable or function.
    """
    output_feature = copy.deepcopy(feature)
    # attributes
    feature_attributes = output_feature.get("attributes", {})
    if feature_attributes:
        # recast field if necessary
        for field_name, field_name_recast in recast_field_mapping.items():
            feature_attributes[field_name] = recast_field_value(
                field_value=feature_attributes[field_name],
                recast_value_to_python_type=field_name_recast[
                    "recast_value_to_python_type"
                ],
                # recast_value_in_list=field_name_recast['recast_value_in_list'],
                resize_value_width=field_name_recast["resize_value_width"],
                resize_value_precision=field_name_recast["resize_value_precision"],
            )
        # reformat data
        if shapefile_format_value:
            for field_name, shp_field_metadata in shapefile_fields_metadata.items():
                reformat_dict_for_field_type = shapefile_format_value.get(
                    shp_field_metadata.get("fieldType")
                )
                if reformat_dict_for_field_type:
                    # get reformat_function
                    reformat_function = _get_function(
                        reformat_dict_for_field_type, locals_var=locals_var
                    )
                    # init parameters
                    parameters = reformat_dict_for_field_type["default_parameters"]
                    # add value to reformat in parameters
                    parameters[
                        reformat_dict_for_field_type["need_parameters_name"]
                    ] = feature_attributes[field_name]
                    # rewrite in feature data with reformat function
                    feature_attributes[field_name] = reformat_function(**parameters)

    return output_feature


def _esri_shapefile_format_field_metadata(
    _geolayer, _driver_parameters, _recast_field_mapping
):
    """
    Return field metadata for output format.

    Indeed, some geoformat field types are not necessarily compatible with the output driver
    (so they have to be recast). Also, the type of fields in the output format do not always
     require the same metadata (we must therefore sort them).

    This function return a dictionary indicating for each field the type in the output format
    and the required metadata with it.

    :param _geolayer: geolayer.
    :param _driver_parameters: driver parameters.
    :param _recast_field_mapping: dict that contains field to recast and parameters for recasting.
    :return: shapefile field metadata
    """
    # make dictionary with output format field type correspondance metadata in it
    shapefile_fields_metadata = {}
    default_field = {
        "id": {"type": "Integer"}
    }  # TODO GET shapefile_driver_function_dict
    geoformat_fields_metadata = _geolayer["metadata"].get("fields", default_field)
    for field_name in geoformat_fields_metadata:
        geoformat_field_type = geoformat_fields_metadata[field_name]["type"]
        shapefile_field_type = geoformat_field_type_to_shapefile_field_type.get(
            geoformat_field_type
        )
        shapefile_field_size = geoformat_fields_metadata[field_name].get("width", None)
        shapefile_field_decimal = geoformat_fields_metadata[field_name].get(
            "precision", None
        )
        if field_name in _recast_field_mapping:
            if _recast_field_mapping[field_name]["recast_value_to_output_type"]:
                shapefile_field_type = _recast_field_mapping[field_name][
                    "recast_value_to_output_type"
                ]
            if _recast_field_mapping[field_name]["resize_value_width"]:
                shapefile_field_size = _recast_field_mapping[field_name][
                    "resize_value_width"
                ]
            if _recast_field_mapping[field_name]["resize_value_precision"]:
                shapefile_field_decimal = _recast_field_mapping[field_name][
                    "resize_value_precision"
                ]
        # add + 1 size for number and float type
        if shapefile_field_type in {"N", "F"}:
            if (
                shapefile_field_decimal is not None and shapefile_field_decimal > 0
            ):  # if there is number after comma
                shapefile_field_size += 1
        # TODO big refacto here
        # TODO is it overkill ? Response : YES
        shapefile_matching_field_variable_dict = {
            0: "name",
            1: "fieldType",
            2: "size",
            3: "decimal",
        }
        field_metadata_list = [
            field_name,
            shapefile_field_type,
            shapefile_field_size,
            shapefile_field_decimal,
        ]
        shapefile_field_variable = {}
        for i_metadata, field_metadata in enumerate(field_metadata_list):
            if field_metadata:
                shapefile_field_variable[
                    shapefile_matching_field_variable_dict[i_metadata]
                ] = field_metadata
        # save shapefile field metadata
        shapefile_fields_metadata[field_name] = shapefile_field_variable

    return shapefile_fields_metadata


def _write_layer(
    _geolayer,
    _output_path,
    _driver_parameters,
    _esri_shapefile_field_metadata,
    _output_field_recast_mapping,
    encoding,
):
    """
    From geolayer to esri shapefile files.

    :param _geolayer: geolayer that will be converted to a shapefile.
    :param _output_path: path for output file
    :param _driver_parameters: esri shapefile driver parameters
    :param _esri_shapefile_field_metadata: esri shapefile field metadata
    :param _output_field_recast_mapping:  dict that contains field name to recast and parameters for recasting
    :param encoding: output encoding string wanted
    """
    # CREATE LAYER
    # create_layer_metadata_dict = driver_function_dict['create_layer']  # TODO DELELE
    create_layer_metadata_dict = _driver_parameters["layer_creator"]["create_layer"]
    create_layer_function = _get_function(create_layer_metadata_dict)

    # fill need parameters
    create_layer_parameters = {}
    layer_creator_need_parameter = create_layer_metadata_dict.get("need_parameters")
    if layer_creator_need_parameter:
        for need_parameter_name, key in create_layer_metadata_dict[
            "need_parameters"
        ].items():
            if key == "PATH":
                key_value = _output_path
                # if no geometry in geolayer we only create a dbf
                if _geolayer["metadata"].get("geometry_ref") is None:
                    need_parameter_name = "dbf"

            elif key == "OUTPUT_GEOMETRY_TYPE":
                geolayer_geometry_ref = _geolayer["metadata"].get("geometry_ref")
                if geolayer_geometry_ref:
                    geolayer_geometry_type = geolayer_geometry_ref["type"]
                    geometry_type_set = set()
                    for geometry_type in geolayer_geometry_type:
                        output_geometry_type = _driver_parameters["driver_formating"][
                            "geometry_type"
                        ][geometry_type]
                        geometry_type_set.update([output_geometry_type])
                    if len(geometry_type_set) == 1:
                        key_value = list(geometry_type_set)[0]
                    else:
                        raise Exception(
                            geometry_type_does_not_allowed_in_geolayer.format(
                                geometry_type=geometry_type_set
                            )
                        )
                else:
                    continue  # if no geometry_ref in geolayer we create only the dbf
                    # then we do not need geometry_type parameter
            else:
                raise Exception(key_parameter_invalid.format(key=key))

            create_layer_parameters[need_parameter_name] = key_value
    create_layer_parameters = {
        **create_layer_metadata_dict["default_parameters"],
        **create_layer_parameters,
    }

    with create_layer_function(
        **create_layer_parameters, encoding=encoding
    ) as create_layer:

        # get method and parameters for field creation
        # get parameters
        field_creator_dict = _driver_parameters["layer_creator"]["create_field"]
        # get method
        field_creator_method = _get_function(field_creator_dict, locals_var=locals())

        # get parameters and associate method for feature creation
        # get parameters
        create_feature_attributes_dict = _driver_parameters["layer_creator"][
            "create_feature_attributes"
        ]
        # get method
        create_feature_attributes_method = _get_function(
            create_feature_attributes_dict, locals_var=locals()
        )
        # CREATE FIELD
        # test if there is mother class to create field
        mother_class_for_create_field = _driver_parameters["layer_creator"][
            "create_field"
        ].get("method_child_from")
        if mother_class_for_create_field:
            # reformat field name (if necessary)
            ori_field_name_list = get_field_name_list_ordered_by_i_field(
                fields_metadata=_esri_shapefile_field_metadata
            )
            max_field_name_char = _driver_parameters["driver_formating"].get(
                "max_field_name_char"
            )
            new_field_name_list = _esri_shapefile_rename_field_list(
                field_name_list=ori_field_name_list,
                max_field_name_char=max_field_name_char,
            )

            for i_field, new_field_name in enumerate(new_field_name_list):
                ori_field_name = ori_field_name_list[i_field]
                shp_field_metadata = _esri_shapefile_field_metadata[ori_field_name]
                shp_field_metadata["name"] = new_field_name

                field_creator_method(**shp_field_metadata)

        # CREATE FEATURE
        # get geolayer fields metadata
        fields_metadata = _geolayer["metadata"].get("fields", None)
        for i_feat, feature in _geolayer["features"].items():

            shapefile_feature = _geoformat_feature_to_shapefile_feature(
                feature=feature,
                shapefile_fields_metadata=_esri_shapefile_field_metadata,
                recast_field_mapping=_output_field_recast_mapping,
                shapefile_format_value=_driver_parameters["driver_formating"][
                    "format_feature_data"
                ],
                locals_var=locals(),
            )
            if shapefile_feature.get("attributes"):
                feature_shapefile_values = [
                    shapefile_feature["attributes"].get(field_name, None)
                    for field_name in fields_metadata
                ]

                # create_feature_attributes_method = getattr(layer, create_feature_attributes_method_name)
                create_feature_attributes_method(*feature_shapefile_values)
            else:
                create_layer.record(i_feat)

            if create_layer.shapeType is not None:  # if not only dbf file
                if feature_geometry := shapefile_feature.get("geometry"):
                    feature_geometry_type = feature_geometry.get("type")
                    feature_geometry_coordinates = feature_geometry.get("coordinates")
                    if not feature_geometry_coordinates:
                        create_layer.null()
                    else:
                        if feature_geometry_type == "Point":
                            create_feature_geometry = create_layer.point
                        elif feature_geometry_type in {"LineString", "MultiLineString"}:
                            create_feature_geometry = create_layer.line
                        elif feature_geometry_type in {"Polygon", "MultiPolygon"}:
                            create_feature_geometry = create_layer.poly
                            # use force rhr
                            feature_geometry = force_rhr(
                                polygon_geometry=feature_geometry
                            )
                        elif feature_geometry_type == "MultiPoint":
                            create_feature_geometry = create_layer.multipoint
                        # write geometry in feature
                        create_feature_geometry(
                            **_format_coordinates_for_pyshp(geometry=feature_geometry)
                        )
                else:
                    create_layer.null()


def geolayer_to_shapefile(
    geolayer,
    path,
    overwrite=False,
    encoding="utf8",
):
    """
    Convert geolayer to esri shapefile file.

    :param geolayer: input geolayer to convert.
    :param path: output file or directory path.
    :param overwrite: True we overwrite if exists false we raise an exception.
    :param encoding: string encoding default utf8
    """
    if import_shapefile_success is False:
        raise Exception(import_shapefile_error)

    # get shapefile driver dict
    driver_parameters = shapefile_driver_function_dict

    # check fields metadata and identifies the fields that must be recast
    output_field_recast_mapping = _esri_shapefile_check_fields_metadata(
        _geolayer=geolayer, _driver_parameters=driver_parameters
    )

    # get adjusted field metadata
    esri_shapefile_field_metadata = _esri_shapefile_format_field_metadata(
        _geolayer=geolayer,
        _driver_parameters=driver_parameters,
        _recast_field_mapping=output_field_recast_mapping,
    )

    # check geometry compatibility
    geometry_metadata = geolayer["metadata"].get("geometry_ref")
    geometry_mapping = driver_parameters["driver_formating"].get("geometry_mapping")
    if geometry_metadata and geometry_mapping:
        geolayer_list = list(
            split_geolayer_by_geometry_type(
                geolayer=geolayer, geometry_type_mapping=geometry_mapping
            )
        )
    else:
        geolayer_list = [geolayer]

    # we have multi geolayer to create then path must be a folder not a file
    if len(geolayer_list) > 1:
        verify_input_path_is_dir(path=path)

    for i_geolayer, _geolayer in enumerate(geolayer_list):
        if _geolayer["metadata"].get("geometry_ref"):
            add_extension = ".shp"
        else:
            add_extension = ".dbf"

        # verify output path
        output_path = path_to_file_path(
            path=path,
            geolayer_name=_geolayer["metadata"]["name"],
            overwrite=overwrite,
            add_extension=add_extension,
        )

        # write layer
        _write_layer(
            _geolayer=_geolayer,
            _output_path=output_path,
            _driver_parameters=driver_parameters,
            _esri_shapefile_field_metadata=esri_shapefile_field_metadata,
            _output_field_recast_mapping=output_field_recast_mapping,
            encoding=encoding,
        )


def _shapefile_fields_to_geoformat_field_metadata(
    shp_fields, shp_rename_field_name_list
):
    """
    Deduce geoformat field metadata for shapefile field parameters

    :param shp_fields: shapefile field object
    :param shp_rename_field_name_list: list that contains output field name
    :return: geoformat geolayer field metadata
    """
    geoformat_field_metadata = {}
    field_index = 0

    # loop on each field and create geolayer metadata
    for i_field, shp_metadata_field in enumerate(shp_fields):
        (
            shp_ori_field_name,
            shp_field_type,
            shp_width,
            shp_precision,
        ) = shp_metadata_field
        shp_field_name = shp_rename_field_name_list[i_field]
        if shp_ori_field_name != "DeletionFlag":
            geoformat_field_type = shapefile_field_type_to_geoformat_field_type.get(
                shp_field_type
            )
            if geoformat_field_type == "Real":
                if shp_precision == 0:
                    geoformat_field_type = "Integer"
                else:
                    shp_width -= 1  # esri shapefile count real comma in width
            geoformat_field_metadata[shp_field_name] = {"type": geoformat_field_type}
            if geoformat_field_type in field_metadata_width_required:
                geoformat_field_metadata[shp_field_name]["width"] = shp_width
            if geoformat_field_type in field_metadata_precision_required:
                geoformat_field_metadata[shp_field_name]["precision"] = shp_precision

            geoformat_field_metadata[shp_field_name]["index"] = field_index
            field_index += 1

    return geoformat_field_metadata


def _get_record_to_layer(layer, feature_offset=None, feature_limit=None):
    """
    Yield record (feature) of layer. You can add offset and limit parameters to have only some record and not all
     layer's records.

    :param layer: shapefile layer object
    :param feature_offset: says to skip that many records before beginning to return record
    :param feature_limit: says the maximum number of records to yield
    :yield: record
    """
    # if just dbf in input
    if layer.shp is None:
        layer = layer.records()

    yield_count = 0
    for i_feat, record in enumerate(layer):
        if feature_offset:
            if i_feat < feature_offset:
                continue

        if record:
            # if just dbf we create a ShapeRecord object for record
            if not isinstance(record, shapefile.ShapeRecord):
                record = shapefile.ShapeRecord(record=record)
            yield record
            yield_count += 1

            if feature_limit:
                if yield_count == feature_limit:
                    break


def _shapefile_record_to_geoformat_feature(
    record,
    selected_field_shapefile_index,
    ordered_field_name_list,
    layer_shape_type,
    pyshp_geometry_type,
    bbox_extent,
):
    """
    Convert shapefile record (feature) to geoformat feature

    :param record: shapefile record
    :param selected_field_shapefile_index: list of field index in
    :param ordered_field_name_list: field name list
    :param layer_shape_type: shapefile geometry type
    :param pyshp_geometry_type: correspondance table between pyshp geometry number and str geometry type
    :param bbox_extent: True we add bbox to feature/ False we don't
    """
    feature = {}
    # attributes
    if record.record and selected_field_shapefile_index is not None:
        feature["attributes"] = {}
        for i_field, shp_i_field in enumerate(selected_field_shapefile_index):
            field_value = record.record[shp_i_field]
            feature["attributes"][ordered_field_name_list[i_field]] = field_value

    # geometry
    if layer_shape_type:
        if record.shape:
            if record.shape.shapeType:
                feature_geometry = record.shape.__geo_interface__
                feature_geometry["coordinates"] = format_coordinates(
                    coordinates_list_tuple=feature_geometry["coordinates"],
                    format_to_type=list,
                )
            else:
                feature_geometry = {
                    "type": pyshp_geometry_type[layer_shape_type],
                    "coordinates": [],
                }

            # add bbox (if required)
            if bbox_extent is True and feature_geometry["coordinates"]:
                if record.shape.shapeType != 1:
                    bbox = tuple(record.shape.bbox)
                else:
                    bbox = geometry_to_bbox(geometry=feature_geometry)
                feature_geometry["bbox"] = bbox

            if feature_geometry:
                # add geometry to feature
                feature["geometry"] = feature_geometry

    return feature


def shapefile_to_geolayer(
    path,
    field_name_filter=None,
    rename_field_dict=None,
    bbox_extent=True,
    bbox_filter=None,
    serialize=False,
    geometry_type_filter=None,
    feature_limit=None,
    feature_offset=None,
    force_field_conversion=False,
    crs=None,
    encoding="utf8",
):
    """
    Convert esri shapefile file to geolayer

    :param path: path to geojson file
    :param field_name_filter: field_name that we want to keep in geolayer (can be a list).
    :param rename_field_dict: dict as table of correspondance to rename field.
    :param bbox_extent: add "bbox" key in each feature and "extent" key in geometry metadata.
    :param bbox_filter: keep only feature that intersects bbox (can be a list of bbox).
    :param serialize: True if features in geolayer are serialized (can reduce performance) / False if not.
    :param geometry_type_filter: keep only features with geometry type in this variable (can be a list).
    :param feature_limit: constrains the number of rows returned in output geolayer.
    :param feature_offset: skip feature before beginning to given line number.
    :param force_field_conversion:  True if you want to force value in field (can change field type) / False if you want
           to deduce field type without forcing field type.
    :param crs: epsg code for coordinates reference system.
    :param encoding: string encoding default utf8
    """

    if import_shapefile_success is False:
        raise Exception(import_shapefile_error)

    field_name_filter = value_to_iterable_value(
        field_name_filter, output_iterable_type=list
    )

    with shapefile.Reader(path, encoding=encoding) as shp:
        geolayer_name = Path(shp.shapeName).name
        output_geolayer = {"metadata": {"name": geolayer_name}, "features": {}}
        # get fields metadata
        # rename field name if necessary
        shp_field_obj = shp.fields
        shp_field_name_list = [
            shp_metadata_field[0] for shp_metadata_field in shp_field_obj
        ]
        shp_rename_field_name_list = _esri_shapefile_rename_field_list(
            field_name_list=shp_field_name_list, max_field_name_char=None
        )
        geolayer_metadata_fields = _shapefile_fields_to_geoformat_field_metadata(
            shp_fields=shp_field_obj,
            shp_rename_field_name_list=shp_rename_field_name_list,
        )
        shapefile_ordered_field_name_list = get_field_name_list_ordered_by_i_field(
            fields_metadata=geolayer_metadata_fields
        )
        if geolayer_metadata_fields:
            output_geolayer["metadata"]["fields"] = geolayer_metadata_fields

        if field_name_filter is not None:
            output_geolayer["metadata"] = drop_field_that_not_exists_in_metadata(
                metadata=output_geolayer["metadata"],
                not_deleting_field_name_or_field_name_list=field_name_filter,
            )

        geolayer_metadata_fields = output_geolayer["metadata"].get("fields")

        # re order fields
        ordered_field_name_list = None
        selected_field_shapefile_index = None
        if geolayer_metadata_fields:
            # create correspondence table between field order in shapefile and kipped field in geolayer
            ordered_field_name_list = get_field_name_list_ordered_by_i_field(
                fields_metadata=output_geolayer["metadata"]["fields"]
            )
            selected_field_shapefile_index = [
                shapefile_ordered_field_name_list.index(field_name)
                for field_name in geolayer_metadata_fields
            ]

        # get geometry metadata
        layer_shape_type = None
        if shp.shp is not None:
            layer_shape_type = shp.shapeType
            if layer_shape_type != 0:
                output_geolayer["metadata"]["geometry_ref"] = {"type": set()}
                if bbox_extent is True:
                    output_geolayer["metadata"]["geometry_ref"]["extent"] = tuple(
                        shp.bbox
                    )

        i_feat = 0
        for record in _get_record_to_layer(
            layer=shp, feature_limit=feature_limit, feature_offset=feature_offset
        ):
            feature = _shapefile_record_to_geoformat_feature(
                record=record,
                selected_field_shapefile_index=selected_field_shapefile_index,
                ordered_field_name_list=ordered_field_name_list,
                layer_shape_type=layer_shape_type,
                pyshp_geometry_type=pyshp_geometry_type,
                bbox_extent=bbox_extent,
            )
            feature = feature_filter(
                feature=feature,
                serialize=serialize,
                field_name_filter=field_name_filter,
                geometry_type_filter=geometry_type_filter,
                bbox_filter=bbox_filter,
                bbox=bbox_extent,
            )
            if feature:
                output_geolayer["features"][i_feat] = feature
                # add metadata geolayer
                feature_geometry = feature.get("geometry")
                if feature_geometry:
                    output_geolayer["metadata"]["geometry_ref"]["type"].update(
                        [feature_geometry["type"]]
                    )
                    if bbox_filter or geometry_type_filter:
                        feature_geometry_bbox = geometry_to_bbox(feature_geometry)
                        if i_feat == 0:
                            geolayer["geometry_ref"]["extent"] = feature_geometry_bbox
                        else:
                            geolayer["geometry_ref"]["extent"] = bbox_union(
                                feature_geometry_bbox,
                                geolayer["geometry_ref"]["extent"],
                            )
                i_feat += 1

        if force_field_conversion is True:
            output_geolayer = feature_list_to_geolayer(
                feature_list=[
                    feature for i_feat, feature in output_geolayer["features"].items()
                ],
                geolayer_name=output_geolayer["metadata"]["name"],
                field_name_filter=None,
                force_field_conversion=force_field_conversion,
                geometry_type_filter=None,
                bbox_filter=None,
                bbox_extent=bbox_extent,
                crs=crs,
                serialize=serialize,
                none_value_pattern=none_value_pattern,
            )

        if rename_field_dict:
            for field_name, new_field_name in rename_field_dict.items():
                output_geolayer = rename_field(
                    geolayer=output_geolayer,
                    old_field_name=field_name,
                    new_field_name=new_field_name,
                )

    return output_geolayer


def _esri_shapefile_rename_field_list(field_name_list, max_field_name_char=None):
    """
    This function rename field_name contains in a list.
    The return field has a length less than or equal to the variable max_field_name_char

    :param field_name_list: list of field to rename
    :param max_field_name_char: maximum number of characters
    :return: list with renamed field_name
    """
    new_field_name_set = set()
    result_field_list = [None] * len(field_name_list)
    for i_field, field_name in enumerate(field_name_list):
        if max_field_name_char:
            new_field_name = field_name[:max_field_name_char]
            new_max_field_length = max_field_name_char
        else:
            new_field_name = field_name
            new_max_field_length = len(field_name)

        if new_field_name in new_field_name_set:
            exists = True
            it = 1
            new_field_name = field_name
            while exists is True:
                len_it = len(str(it))
                max_field_lenght_with_length_it = new_max_field_length - len_it
                if (
                    max_field_lenght_with_length_it < 1
                ):  # alway keep the first letter of input field_name
                    raise Exception(
                        field_name_length_not_valid.format(
                            variable_name=max_field_name_char
                        )
                    )
                new_field_name = new_field_name[:max_field_lenght_with_length_it]

                new_field_name = f"{new_field_name}{it}"
                if new_field_name in new_field_name_set:
                    it += 1
                else:
                    exists = False

        result_field_list[i_field] = new_field_name
        new_field_name_set.update([new_field_name])

    return result_field_list



def create_attribute_index(geolayer, field_name):
    index_dict = {"metadata": {'type': 'hashtable'}, 'index': {}}

    # récupération de la valeur du champs à indexer
    for i_feat in geolayer['features']:
        feature = geolayer['features'][i_feat]

        # if feature is serialized
        if 'feature_serialize' in geolayer['metadata']:
            if geolayer['metadata']['feature_serialize']:
                feature = eval(feature)

        field_value = feature['attributes'][field_name]

        try:
            index_dict['index'][field_value].append(i_feat)
        except KeyError:
            index_dict['index'][field_value] = [i_feat]

    return index_dict



import hashlib

from geoformat.conf.error_messages import (
    variables_must_be_true,
    variable_must_be_fill,
)
from geoformat.conversion.geometry_conversion import  geometry_to_wkb
from geoformat.conversion.metadata_conversion import  get_field_name_list_ordered_by_i_field


def get_feature_attributes_hash(feature_attributes, field_attributes_field_name_order):
    """
    return sha256 of feature attributes.
    We order field in attributes with field_attributes_field_name_order as specified in geolayer
    field order metadata. -> We do that to compare feature attributes with None value (because when a value is None in
    a feature we are not obliged to add field with None value in it).

    :param feature_attributes: feature attributes to hash
    :param field_attributes_field_name_order: list/tuple of field name order
    :return: sha256 or feature attributes
    """
    # if feature_attributes and field_attributes_field_name_order:
    if field_attributes_field_name_order:
        if feature_attributes is None:
            feature_attributes = {}
        ordered_value_list = [None] * len(field_attributes_field_name_order)
        for i_field, field_name in enumerate(field_attributes_field_name_order):
            ordered_value_list[i_field] = (field_name, feature_attributes.get(field_name))
    else:
        ordered_value_list = []

    return hashlib.sha256(str(ordered_value_list).encode('utf8')).hexdigest()


def get_feature_geometry_hash(feature_geometry):
    """
    return sha256 of feature geometry.
    feature geometry is convert in wkb, and we make a hash of this.

    :param feature_geometry: feature geometry
    :return: sha256 or feature geometry
    """
    if feature_geometry:
        geometry_to_hash = geometry_to_wkb(feature_geometry)
    else:
        geometry_to_hash = ''.encode('utf8')

    return hashlib.sha256(geometry_to_hash).hexdigest()


def get_feature_hash(feature, field_attributes_field_name_order=None, attribute_hash=True, geometry_hash=True):
    """
    return sha 256 of feature.

    :param feature: feature to hash
    :param field_attributes_field_name_order: list/tuple of field name order
    :param attribute_hash: True if you want attribute hash
    :param geometry_hash: True if you want geometry hash
    :return: sha256 of feature
    """
    if attribute_hash is False and geometry_hash is False:
        raise Exception(variables_must_be_true.format(variables=' or '.join(['attribute_hash', 'geometry_hash'])))
    if attribute_hash is True and feature.get('attributes') and not field_attributes_field_name_order:
        raise Exception(variable_must_be_fill.format(variable_name="field_attributes_field_name_order"))

    feature_attributes_hash = ''
    feature_geometry_hash = ''
    feature_attributes = feature.get("attributes")
    if attribute_hash is True and feature_attributes and field_attributes_field_name_order:
        feature_attributes_hash = get_feature_attributes_hash(feature_attributes=feature_attributes,
                                                              field_attributes_field_name_order=field_attributes_field_name_order)
    feature_geometry = feature.get("geometry")
    if geometry_hash is True and feature_geometry:
        feature_geometry_hash = get_feature_geometry_hash(feature_geometry=feature_geometry)

    feature_hash = hashlib.sha256((feature_attributes_hash + feature_geometry_hash).encode('utf8')).hexdigest()

    return feature_hash


def get_duplicate_features(geolayer,  check_attribute_duplicate=True, check_geometry_duplicate=True):
    """
    Scan geolayer and return duplicate feature.
    You can optionally choose if you want to determine duplication, resume by this table
        | geometry_duplicate |attribute_duplicate | meaning                  |
        |--------------------|--------------------|--------------------------|
        | True               | True               | geometry and attributes  |
        | True               | False              | geometry only            |
        | False              | True               | attributes only          |
        | False              | False              | Error                    |

    :param geolayer: input geolayer for which we want to know the duplicated entities
    :param check_attribute_duplicate: True if you want to check attributes
    :param check_geometry_duplicate: True if you want check geometry
    :yield: duplicate feature i_feat
    """

    if check_geometry_duplicate is False and check_attribute_duplicate is False:
        raise Exception(variables_must_be_true.format(variables=' or '.join(['geometry_duplicate', 'attribute_duplicate'])))

    # initialize checking variable
    field_attributes_field_name_order = {}
    if geolayer['metadata'].get('fields'):
        field_attributes_field_name_order = get_field_name_list_ordered_by_i_field(fields_metadata=geolayer['metadata']['fields'])

    duplicate_hash = set()
    for i_feat, feature in geolayer['features'].items():
        feature_hash = get_feature_hash(feature=feature, field_attributes_field_name_order=field_attributes_field_name_order, attribute_hash=check_attribute_duplicate, geometry_hash=check_geometry_duplicate)

        if feature_hash in duplicate_hash:
            yield i_feat

        duplicate_hash.update([feature_hash])



proj_correspondance = {
   "CRS84": 4326,         # (WGS 84 / Auto UTM )
   "CRS83": 2954,         # (NAD83 longitude-latitude)
   "CRS27": 4267,         # (NAD27 longitude-latitude)
   "NAD83": 2954,         # (NAD83 longitude-latitude)
   "NAD27": 4267,         # (NAD27 longitude-latitude)
   "AUTO2:42001": None,   # (WGS 84 / Auto UTM )
   "AUTO2:42002": 9807,   # (WGS 84 / Auto Tr. Mercator)
   "AUTO2:42003": 9840,   # (WGS 84 / Auto Orthographic)
   "AUTO2:42004": 4087,   # (WGS 84 / Auto Equirectangular)
   "AUTO2:42005": 54009,  # (WGS 84 / Auto Mollweide)
   "AUTO2:97001": None,   # Gnonomic
   "AUTO2:97002": 54026,  # Stereographic
   "AUTO2:97003": 54032,  # Azimuthal Equidistant
}


import copy

from geoformat.conf.error_messages import (
    field_exists,
    metadata_geometry_ref_not_found,
    metadata_geometry_ref_type_not_match,
    geometry_ref_metadata,
    geometry_ref_feature
)
from geoformat.conf.format_data import (
    value_to_iterable_value
)
from geoformat.manipulation.feature_manipulation import (
    rename_field_in_feature,
    drop_field_that_not_exists_in_feature
)
from geoformat.manipulation.metadata_manipulation import (
    drop_field_that_not_exists_in_metadata,
    rename_field_in_metadata
)


def _merge_metadata_rename_field(metadata, field_to_rename_dict, rename_field_correspondance):
    """
    function build for merge_metadata
    """
    for i_field, (old_field_name, new_field_name) in enumerate(field_to_rename_dict.items()):
        metadata = rename_field_in_metadata(
            metadata=metadata,
            old_field_name=old_field_name,
            new_field_name=new_field_name
        )
        if i_field == 0:
            rename_field_correspondance = {}
        rename_field_correspondance[old_field_name] = new_field_name

    return metadata, rename_field_correspondance


def merge_metadata(
        metadata_a,
        metadata_b,
        geolayer_name,
        field_name_filter_a=None,
        field_name_filter_b=None,
        geometry_ref="metadata_a",
        rename_output_field_from_geolayer_a="auto",
        rename_output_field_from_geolayer_b="auto",
):
    """
    When a join is performed between two geolayers : it results a new geolayer whose attribute component is the result
    of the merging of fields metadata.
    You can optionally choose here to keep the geometry metadata of this or that geolayer.

    In output we receive not only a new metadata but also a fields correspondance table : in case fields with the same
    name exist between the two geolayers a mapping table is provided as output.

    :param metadata_a: first metadata to merge
    :param metadata_b: second metadata to merge
    :param geolayer_name: new name of geolayer
    :param field_name_filter_a: field name filter for metadata_a
    :param field_name_filter_b: field name filter for metadata_b
    :param geometry_ref: name the metadata that keep the geometry_ref metadata (default : metadata_a)
    :param rename_geolayer_field: indicate when there is name field collision what we do. Two option :
                                - "auto" : rename automatiquely the field
                                - a dict with this format : {
                                "metadata_a":
                                    {old_field_name_a: new_field_name_a,
                                     old_filed_name_b: new_field_name_b ...},
                                "metadata_b:
                                    {old_field_name_c: new_field_name_c,
                                     old_filed_name_d: new_field_name_d ...},
                                }
    :return: the merged metadata
    """
    metadata_merged = copy.deepcopy(metadata_a)
    metadata_b = copy.deepcopy(metadata_b)
    field_name_filter_a = value_to_iterable_value(value=field_name_filter_a, output_iterable_type=list)
    field_name_filter_b = value_to_iterable_value(value=field_name_filter_b, output_iterable_type=list)

    # delete geometry ref
    if 'geometry_ref' in metadata_merged and geometry_ref == 'metadata_b':
        del metadata_merged['geometry_ref']

    # add new name to merge metadata
    metadata_merged["name"] = geolayer_name

    # clean index and matrix
    if 'index' in metadata_merged:
        del metadata_merged['index']
    if 'matrix' in metadata_merged:
        del metadata_merged['matrix']

    # drop useless field
    if field_name_filter_a is not None and metadata_merged.get('fields'):
        metadata_merged = drop_field_that_not_exists_in_metadata(
            metadata=metadata_merged,
            not_deleting_field_name_or_field_name_list=field_name_filter_a
        )
    if field_name_filter_b is not None and metadata_b.get('fields'):
        metadata_b = drop_field_that_not_exists_in_metadata(
            metadata=metadata_b,
            not_deleting_field_name_or_field_name_list=field_name_filter_b
        )

    # rename field if necessary
    rename_field_correspondance_a = {}
    if rename_output_field_from_geolayer_a != 'auto':
        metadata_merged, rename_field_correspondance_a = _merge_metadata_rename_field(
            metadata=metadata_merged,
            field_to_rename_dict=rename_output_field_from_geolayer_a,
            rename_field_correspondance=rename_field_correspondance_a)
    rename_field_correspondance_b = {}
    if rename_output_field_from_geolayer_b != 'auto':
        metadata_b, rename_field_correspondance_b = _merge_metadata_rename_field(
            metadata=metadata_b,
            field_to_rename_dict=rename_output_field_from_geolayer_b,
            rename_field_correspondance=rename_field_correspondance_b)

    # add fields in metadata
    for key, value in metadata_b.items():
        if key == 'fields':
            for geolayer_b_field_name, geolayer_b_field_metadata in metadata_b[key].items():
                if 'fields' in metadata_merged:
                    # if geolayer_b_field_name exists in geolayer_a
                    if metadata_merged['fields'].get(geolayer_b_field_name):
                        if rename_output_field_from_geolayer_b == 'auto':
                            exists = True
                            it = 1
                            new_field_name = geolayer_b_field_name
                            while exists is True:
                                new_field_name = f'{new_field_name}{it}'
                                if new_field_name in metadata_merged['fields']:
                                    it += 1
                                else:
                                    exists = False
                            # add new_field_name to rename_field_correspondance_b
                            rename_field_correspondance_b[geolayer_b_field_name] = new_field_name
                        else:
                            if geolayer_b_field_name in metadata_merged['fields']:
                                raise Exception(field_exists.format(field_name=geolayer_b_field_name))
                    else:
                        new_field_name = geolayer_b_field_name
                else:
                    # add fields key in metadata
                    metadata_merged['fields'] = {}
                    new_field_name = geolayer_b_field_name

                # create new field in metadata
                metadata_merged['fields'][new_field_name] = dict(geolayer_b_field_metadata)
                # reorder index
                metadata_merged['fields'][new_field_name]['index'] = len(metadata_merged['fields']) - 1

        # update geometry ref if necessary
        if key == 'geometry_ref':
            if geometry_ref == 'metadata_a':
                pass
            elif geometry_ref == 'metadata_b':
                metadata_merged['geometry_ref'] = copy.deepcopy(metadata_b['geometry_ref'])
            else:
                raise Exception(geometry_ref_metadata)

    return {
        'metadata': metadata_merged,
        'fields_correspondance_a': rename_field_correspondance_a,
        'fields_correspondance_b': rename_field_correspondance_b
    }


def merge_feature(
    feature_a,
    feature_b,
    merge_metadata=None,
    field_name_filter_a=None,
    field_name_filter_b=None,
    rename_fields_a=None,
    rename_fields_b=None,
    geometry_ref='feature_a'
):
    """
    Merge two features in one new feature.

    :param feature_a: first feature to merge
    :param feature_b: second feature to merge
    :param merge_metadata: if we want add a constraint metadata to merge. Warning : here metadata are not updated by
    feature that we want to merge. It is a constraint to check if the new feature will be compatible with given metadata
    :param field_name_filter_a:  field name filter for feature_a
    :param field_name_filter_b:  field name filter for feature_b
    :param rename_fields_a: rename field mapping for feature_a
    :param rename_fields_b: rename field mapping for feature_b
    :param geometry_ref: feature that keep geometry references (default value : feature_a)
    :return: merged result of feature_a and feature_bgeo
    """
    feature_a = copy.deepcopy(feature_a)
    feature_b = copy.deepcopy(feature_b)

    # delete useless field
    if field_name_filter_a:
        feature_a = drop_field_that_not_exists_in_feature(
            feature=feature_a,
            not_deleting_field_name_or_field_name_list=field_name_filter_a
        )
    if field_name_filter_b:
        feature_b = drop_field_that_not_exists_in_feature(
            feature=feature_b,
            not_deleting_field_name_or_field_name_list=field_name_filter_b
        )

    # rename fields
    if rename_fields_a:
        for old_field_name, new_field_name in rename_fields_a.items():
            feature_a = rename_field_in_feature(
                feature=feature_a,
                old_field_name=old_field_name,
                new_field_name=new_field_name
            )

    if rename_fields_b:
        for old_field_name, new_field_name in rename_fields_b.items():
            feature_b = rename_field_in_feature(
                feature=feature_b,
                old_field_name=old_field_name,
                new_field_name=new_field_name
            )

    # check there is no field_name collision between feature
    feature_field_name_collision = set(feature_a.get('attributes', {}).keys()).intersection(set(feature_b.get('attributes', {}).keys()))
    if len(feature_field_name_collision) > 0:
        error_message_list = []
        for field_name in feature_field_name_collision:
            error_message_list.append(field_exists.format(field_name=field_name))
        raise Exception('/n'.join(error_message_list))

    # write attributes
    output_feature = {}
    feature_attributes_list = []
    if feature_a_attributes := feature_a.get('attributes'):
        feature_attributes_list.append(feature_a_attributes)
    if feature_b_attributes := feature_b.get('attributes'):
        feature_attributes_list.append(feature_b_attributes)

    # if merge metadata we check fields and keep only matching fields
    if merge_metadata:
        if fields_metadata := merge_metadata.get('fields'):
            for field_name, field_metadata in fields_metadata.items():
                for feature_attributes in feature_attributes_list:
                    if field_name in feature_attributes:
                        if 'attributes' not in output_feature:
                            output_feature['attributes'] = {}
                        output_feature['attributes'][field_name] = feature_attributes[field_name]

    else:
        for feature_attributes in feature_attributes_list:
            for field_name, field_value in feature_attributes.items():
                if 'attributes' not in output_feature:
                    output_feature['attributes'] = {}
                output_feature['attributes'][field_name] = field_value

    # write geometry
    if geometry_ref == 'feature_a':
        output_feature_geometry = feature_a.get('geometry')
    elif geometry_ref == 'feature_b':
        output_feature_geometry = feature_b.get('geometry')
    else:
        raise Exception(geometry_ref_feature)

    # add geometry to feature
    if output_feature_geometry is not None:
        if merge_metadata:
            if metadata_geometry :=merge_metadata.get('geometry_ref'):
                if output_feature_geometry['type'] not in metadata_geometry['type']:
                    raise Exception(metadata_geometry_ref_type_not_match)
            else:
                raise Exception(metadata_geometry_ref_not_found)

        output_feature['geometry'] = output_feature_geometry

    return output_feature



from geoformat.conf.error_messages import non_unique_values
from geoformat.conversion.feature_conversion import feature_deserialize
from geoformat.conf.format_data import value_to_iterable_value
from geoformat.conf.error_messages import field_missing


def create_pk(geolayer, pk_field_name):
    """
    Return pk dictionary

    :param geolayer: geolayer
    :param pk_field_name: field that we want add constraint
    :return: pk dictionary
    """
    # chek if field exist in geolayer
    field_exists = False
    if 'fields' in geolayer['metadata']:
        field_exists = geolayer['metadata']['fields'].get(pk_field_name, False)

    if field_exists:
        # create empty dictionary
        pk_dictionary = {}

        # get value to index
        for i_feat in geolayer['features']:
            feature = geolayer['features'][i_feat]

            # deserialize feature if serialized
            if 'feature_serialize' in geolayer['metadata']:
                if geolayer['metadata']['feature_serialize']:
                    feature = feature_deserialize(serialized_feature=feature, bbox=False)

            pk_field_value = feature['attributes'][pk_field_name]

            # if pk_field_value is list we have to transform to tuple
            if isinstance(pk_field_value, list):
                pk_field_value = value_to_iterable_value(value=pk_field_value, output_iterable_type=tuple)

            # check if value is unique or not
            if pk_field_value in pk_dictionary:
                raise Exception(non_unique_values.format(field_name=pk_field_name))
            else:
                # save i_feat in pk dictionary
                pk_dictionary[pk_field_value] = i_feat

        return pk_dictionary
    else:
        raise Exception(field_missing.format(field_name=pk_field_name))



import random

def random_point(bbox, nb_round=2):
    """
    Renvoie un point au hasard compris dans une bbox

    Entree :
        bbox : boundary box dans laquelle on veut voir le point

    Sortie :
        Retourne un point au hasard dans la bbox
    """

    (min_x, min_y, max_x, max_y) = bbox

    x = round(min_x + random.random() * (max_x - min_x), nb_round)
    y = round(min_y + random.random() * (max_y - min_y), nb_round)

    return x, y


def random_segment(bbox, nb_round=2):
    """

    :param bbox:
    :param nbRound:
    :return:
    """
    (min_x, min_y, max_x, max_y) = bbox

    xBegin = round(min_x + random.random() * (max_x - min_x), nb_round)
    yBegin = round(min_y + random.random() * (max_y - min_y), nb_round)
    xEnd = round(min_x + random.random() * (max_x - min_x), nb_round)
    yEnd = round(min_y + random.random() * (max_y - min_y), nb_round)

    return (xBegin, yBegin), (xEnd, yEnd)


def random_bbox(bbox, nb_round=2):
    """
    Return random bbox

        Input:
            bbox

        Output:
            new_bbox (bbox)
    """
    x_list, y_list = [0, 0], [0, 0]
    for i in range(2):
        x_list[i] = round(bbox[0] + random.random() * (bbox[2] - bbox[0]), nb_round)
        y_list[i] = round(bbox[1] + random.random() * (bbox[3] - bbox[1]), nb_round)

    new_bbox = (min(x_list), min(y_list), max(x_list), max(y_list))

    return new_bbox




from geoformat.geoprocessing.connectors.operations import coordinates_to_point
from geoformat.conversion.geometry_conversion import geometry_to_geometry_collection


def len_coordinates(coordinates):
    """
    Return number of coordinates on a coordinates list
    """

    coordinates_count = 0
    for point in coordinates_to_point(coordinates):
        coordinates_count += 1

    return coordinates_count


def len_coordinates_in_geometry(geometry):
    """
    Return number of coordinates on a given geometry
    """
    geometry_collection = geometry_to_geometry_collection(geometry)
    coordinates_count = 0
    for geometry in geometry_collection['geometries']:
        coordinates_count += len_coordinates(geometry['coordinates'])

    return coordinates_count



# python version
python_inferior_to_3_7_forbidden = (
    "you must have a version of python higher or equal to 3.7 to execute this function"
)
python_inferior_to_3_8_forbidden = (
    "you must have a version of python higher or equal to 3.8 to execute this function"
)

# driver
import_lib_error = "cannot run this function ! {lib} library not installed"
import_ogr_error = import_lib_error.format(lib="Python-gdal")
import_psycopg2_error = import_lib_error.format(lib="psycopg2")
import_pyproj_error = import_lib_error.format(lib="pyproj")
import_shapefile_error = import_lib_error.format(lib="shapefile")
import_matplotlib_error = import_lib_error.format(lib="matplotlib")
import_numpy_error = import_lib_error.format(lib="numpy")

key_parameter_invalid = "key : {key} not supported"


# path
path_not_valid = "path : {path} is not valid"
path_not_a_dir = f"{path_not_valid}, it must be a directory"
path_not_a_file = f"{path_not_valid}, it must be a file"
path_not_valid_file_exists_overwrite_is_false = "path : {path} exists."
path_http_not_valid = "{path} not valid, code : {code}"
path_not_http = "{path} is not http path"
path_not_exists = "cannot load data file or http address does not exists"

# fields
field_missing = "field : {field_name} does not exists."
field_exists = "{field_name} still exists"
field_type_not_valid = "field type {field_type} not valid. You must use a type in this list : {field_type_list}"
field_width_not_valid = "width value for field {field_name} must be specified"
field_precision_not_valid = "precision value for field {field_name} must be specified"
field_name_length_not_valid = "the maximum number of characters possible for a field (defined by the variable {variable_name}) does not allow to create a unique field."

# geolayer
geolayer_attributes_missing = "there is no attributes data in {geolayer_name} geolayer"

# values

# non unique value
non_unique_values = "field : {field_name} contains non-unique values"

# geometry format
geometry_format_not_exists = (
    "geometry format does not exists you must choose between GEOJSON, WKB or WKT format"
)
geometry_type_not_allowed = "geometry type : {geometry_type} not allowed"
geometry_type_does_not_allowed_in_geolayer = geometry_type_not_allowed + "in geolayer"
geometry_must_be_polygon_or_multipolygon = 'geometry must be POLYGON or MULTIPOLYGON'

# metadata
metadata_fields_not_same = "metadata fields must be identical"
metadata_geometry_ref_not_found = "there is no geometry referenced in metadata"
metadata_geometry_ref_type_not_match = "geometry type is not compatible with metadata"
metadata_geometry_crs = "crs must be identical"

# variable
variable_wrong_formatting = (
    "variable {variable_name} not well formatted please refer to the doc"
)
variable_input_value_error = "variable {variable_name} input error (see doc)"
variable_must_be_int = "variable {variable_name} must be in int"
variable_must_be_fill = 'variable {variable_name} must be fill'

# join
geometry_ref_geolayer = "geometry_ref value must be 'geolayer_a' or 'geolayer_b'"
geometry_ref_metadata = "geometry_ref value must be 'metadata_a' or 'metadata_b'"
geometry_ref_feature = "geometry_ref value must be 'feature_a' or 'feature_b'"

# index
field_name_not_indexing = "field name : {field_name} not indexing the index type is not the one you are looking for"
field_name_still_indexing = "{field_name} is still indexing"
# date to str format
date_format_error = "value must one/or a list of this key 'year' 'month' 'day'  'hour' 'second' 'microsecond'"

# geometry type
geometry_not_in_variable = "{geometry_type} must be include in {variable_name}"
key_use_only_by_geometry = "{key} can only be used on {geometry_type} type"

# Others
field_width_float_not_none = "field_width_float cannot be None if field_type is 'Float'"
variables_must_be_true = "At least one of this variables ({variables}) must be True"

# feature / i_feat
feature_missing = "feature : {i_feat} does not exists"


"""
Contributors :
  Guilhain Averlant (Maintainer)
  Eliette Catelin
  Quentin Lecuire
  Charlotte Montesinos Chevalley
  Coralie Rabiniaux
"""
from geoformat.conf.format_data import is_hexadecimal

from geoformat.constraints.primary_key import create_pk

from geoformat.conversion.bbox_conversion import (
    envelope_to_bbox,
    bbox_to_envelope,
    bbox_extent_to_2d_bbox_extent,
    bbox_to_polygon_coordinates
)
from geoformat.conversion.bytes_conversion import (
    int_to_4_bytes_integer,
    float_to_double_8_bytes_array,
    coordinates_list_to_bytes,
    double_8_bytes_to_float,
    integer_4_bytes_to_int
)
from geoformat.conversion.coordinates_conversion import (
    format_coordinates,
    coordinates_to_2d_coordinates,
    coordinates_to_centroid,
    separate_coordinates_dimensions,
    force_rhr_polygon_coordinates
)
from geoformat.conversion.datetime_conversion import (
    format_datetime_object_to_str_value,
    date_to_int,
    time_to_int,
    datetime_to_timestamp,
    int_to_date,
    int_to_time,
    timestamp_to_datetime,
)
from geoformat.conversion.feature_conversion import (
    feature_serialize,
    feature_deserialize,
    features_geometry_ref_scan,
    check_if_value_is_from_datetime_lib,
    check_if_value_is_date,
    check_if_value_is_datetime,
    check_if_value_is_time,
    return_if_value_is_time_date_or_datetime,
    features_fields_type_scan,
    feature_filter_geometry,
    feature_filter_attributes,
    feature_filter,
    features_filter
)
from geoformat.conversion.fields_conversion import (
    update_field_index,
    recast_field_value,
    recast_field
)

from geoformat.conversion.geolayer_conversion import (
    multi_geometry_to_single_geometry_geolayer,
    geolayer_to_2d_geolayer,
    create_geolayer_from_i_feat_list,
    reproject_geolayer
)

from geoformat.conversion.geometry_conversion import (
    geometry_type_to_2d_geometry_type,
    geometry_to_2d_geometry,
    geometry_to_geometry_collection,
    single_geometry_to_multi_geometry,
    multi_geometry_to_single_geometry,
    geometry_to_multi_geometry,
    ogr_geometry_to_geometry,
    geometry_to_ogr_geometry,
    geometry_to_wkb,
    wkb_to_geometry,
    geometry_to_wkt,
    wkt_to_geometry,
    force_rhr,
    geometry_to_bbox,
    reproject_geometry
)

from geoformat.conversion.metadata_conversion import (
    get_field_name_list_ordered_by_i_field,
    geometries_scan_to_geometries_metadata,
    fields_scan_to_fields_metadata,
    from_field_scan_determine_field_width
)
from geoformat.manipulation.metadata_manipulation import reorder_metadata_field_index_after_field_drop

from geoformat.conversion.precision_tolerance_conversion import (
    deduce_rounding_value_from_float,
    deduce_precision_from_round
)

from geoformat.conversion.segment_conversion import segment_list_to_linestring

from geoformat.db.db_request import (
    sql,
    sql_select_to_geolayer
)

from geoformat.draw.draw import (
    draw_geometry,
    draw_feature,
    draw_geolayer,
)

from geoformat.driver.ogr.ogr_driver import (
    ogr_layer_to_geolayer,
    ogr_layers_to_geocontainer,
    geolayer_to_ogr_layer,
    geocontainer_to_ogr_data_source,
    geoformat_geom_type_to_ogr_geom_type,
    verify_geom_compatibility
)

from geoformat.driver.csv_driver import (
    csv_to_geolayer,
    geoformat_feature_to_csv_feature,
    feature_attributes_to_csv_attributes,
    geolayer_to_csv
)

from geoformat.driver.esri_shapefile_driver import (
    shapefile_to_geolayer,
    geolayer_to_shapefile
)


from geoformat.driver.geojson_driver import (
    feature_attributes_to_properties,
    geoformat_feature_to_geojson_feature,
    geolayer_to_geojson,
    json_object_to_feature_generator,
    from_geojson_get_features_list,
    geojson_to_geolayer
)

from geoformat.driver.postgresql_driver import (
    fields_metadata_to_create_table_fields_structure,
    geometry_ref_to_geometry_field_structure,
    format_to_posgresq_values,
    geolayer_to_postgres
)
from geoformat.explore_data.duplicate import (
    get_feature_hash,
    get_feature_geometry_hash,
    get_feature_attributes_hash,
    get_duplicate_features
)


from geoformat.explore_data.print_data import (
    print_features_data_table,
    print_metadata_field_table
)

from geoformat.explore_data.random_geometry import (
    random_point,
    random_segment,
    random_bbox
)

from geoformat.geoprocessing.connectors.operations import (
    coordinates_to_point,
    coordinates_to_segment,
    coordinates_to_bbox,
    segment_to_bbox
)

from geoformat.geoprocessing.connectors.predicates import (
    point_intersects_point,
    point_intersects_segment,
    point_intersects_bbox,
    segment_intersects_segment,
    segment_intersects_bbox,
    bbox_intersects_bbox,
    ccw_or_cw_segments,
    point_position_segment
)

from geoformat.geoprocessing.generalization.ramer_douglas_peucker import ramer_douglas_peucker
from geoformat.geoprocessing.generalization.visvalingam_whyatt import visvalingam_whyatt

from geoformat.geoprocessing.geoparameters.bbox import (
    bbox_expand,
    bbox_union,
    point_bbox_position
)

from geoformat.geoprocessing.geoparameters.boundaries import(
    ccw_or_cw_boundary
)

from geoformat.geoprocessing.geoparameters.lines import (
    get_slope_between_two_points,
    get_intercept_from_point_and_slope,
    get_slope_from_point_and_intercept,
    line_parameters,
    perpendicular_line_parameters_at_point,
    point_at_distance_with_line_parameters,
    crossing_point_from_lines_parameters
)

from geoformat.geoprocessing.matrix.adjacency import (
    create_adjacency_matrix,
    get_area_intersecting_neighbors_i_feat,
    get_neighbor_i_feat
)

from geoformat.geoprocessing.measure.mesure_area import shoelace_formula, triangle_area

from geoformat.geoprocessing.measure.mesure_distance import (
    euclidean_distance,
    manhattan_distance,
    euclidean_distance_point_vs_segment,
    point_vs_segment_distance
)

from geoformat.geoprocessing.measure.mesure_length import segment_length

from geoformat.geoprocessing.area import geometry_area

from geoformat.geoprocessing.length import geometry_length

from geoformat.geoprocessing.line_merge import line_merge

from geoformat.geoprocessing.merge_geometries import merge_geometries

from geoformat.geoprocessing.point_on_linestring import (
    point_at_a_distance_on_segment,
    points_on_linestring_distance
)

from geoformat.geoprocessing.simplify import (simplify)

from geoformat.geoprocessing.split import (
    segment_split_by_point,
    linestring_split_by_point
)

from geoformat.geoprocessing.union import union_by_split

from geoformat.index.attributes.hash import create_attribute_index

from geoformat.index.geometry.grid import (
    bbox_to_g_id,
    point_to_g_id,
    g_id_to_bbox,
    g_id_to_point,
    g_id_neighbor_in_grid_index,
    create_grid_index,
    grid_index_to_geolayer
)

from geoformat.manipulation.feature_manipulation import (
    rename_field_in_feature,
    drop_field_in_feature,
    drop_field_that_not_exists_in_feature
)

from geoformat.manipulation.geolayer_manipulation import (
    add_attributes_index,
    check_attributes_index,
    check_if_field_exists,
    delete_attributes_index,
    delete_feature,
    drop_field,
    rename_field,
    create_field,
    feature_list_to_geolayer,
    split_geolayer_by_geometry_type,
)

from geoformat.manipulation.metadata_manipulation import (
    add_attributes_index_in_metadata,
    check_attributes_index_in_metadata,
    check_if_field_exists_in_metadata,
    delete_attributes_index_in_metadata,
    drop_field_in_metadata,
    drop_field_that_not_exists_in_metadata,
    reorder_metadata_field_index_after_field_drop,
    rename_field_in_metadata,
    create_field_in_metadata
)

from geoformat.obj.geometry import (
    len_coordinates,
    len_coordinates_in_geometry
)

from geoformat.processing.data.join.join import (
    join,
    join_left,
    join_right,
    join_full
)

from geoformat.processing.data.join.merge_objects import (
    merge_metadata,
    merge_feature
)

from geoformat.processing.data.clauses import (
    clause_where,
    clause_group_by,
    clause_order_by
)

from geoformat.processing.data.union import (
    union_metadata,
    union_geolayer
)

from geoformat.processing.data.field_statistics import field_statistics

from geoformat._version import __version__

# from test_all import test_all

__all__ = (
    # conf.format_data
        [is_hexadecimal] +
        # constraints
        [create_pk] +
        # conversion.bbox
        [envelope_to_bbox, bbox_to_envelope, bbox_extent_to_2d_bbox_extent,  bbox_to_polygon_coordinates] +
        # conversion.bytes
        [int_to_4_bytes_integer, float_to_double_8_bytes_array, coordinates_list_to_bytes, double_8_bytes_to_float,
     integer_4_bytes_to_int] +
        # conversion.coordinates
        [format_coordinates, coordinates_to_2d_coordinates, coordinates_to_centroid, separate_coordinates_dimensions,
     force_rhr_polygon_coordinates] +
        # conversion.datetime
        [format_datetime_object_to_str_value, date_to_int, time_to_int, datetime_to_timestamp, int_to_date, int_to_time,
    timestamp_to_datetime] +
        # conversion.feature
        [feature_serialize, feature_deserialize, features_geometry_ref_scan, check_if_value_is_from_datetime_lib,
     check_if_value_is_date, check_if_value_is_datetime, check_if_value_is_time,
     return_if_value_is_time_date_or_datetime, features_fields_type_scan, feature_list_to_geolayer,
     feature_filter_geometry, feature_filter_attributes, feature_filter, features_filter] +
        # conversion.fields
        [update_field_index, recast_field_value, recast_field] +
        # conversion.geolayer
        [multi_geometry_to_single_geometry_geolayer, geolayer_to_2d_geolayer, create_geolayer_from_i_feat_list,
     reproject_geolayer] +
        # conversion.geometry
        [geometry_type_to_2d_geometry_type, geometry_to_2d_geometry, geometry_to_geometry_collection,
     single_geometry_to_multi_geometry, multi_geometry_to_single_geometry, geometry_to_multi_geometry,
     ogr_geometry_to_geometry, geometry_to_ogr_geometry, geometry_to_wkb, wkb_to_geometry, geometry_to_wkt,
     wkt_to_geometry, force_rhr, geometry_to_bbox, reproject_geometry] +
        # conversion.metadata
        [get_field_name_list_ordered_by_i_field, geometries_scan_to_geometries_metadata, fields_scan_to_fields_metadata,
     reorder_metadata_field_index_after_field_drop, from_field_scan_determine_field_width] +
        # conversion.precision_tolerance
        [deduce_rounding_value_from_float, deduce_precision_from_round] +
        # conversion.segment_conversion
        [segment_list_to_linestring] +
        # draw
        [draw_geometry, draw_feature, draw_geolayer] +
        # db
        [sql, sql_select_to_geolayer] +
        # driver.ogr
        [ogr_layer_to_geolayer, ogr_layers_to_geocontainer, geolayer_to_ogr_layer, geocontainer_to_ogr_data_source,
     geoformat_geom_type_to_ogr_geom_type, verify_geom_compatibility] +
        # driver.csv
        [csv_to_geolayer, geoformat_feature_to_csv_feature, feature_attributes_to_csv_attributes,
     feature_attributes_to_csv_attributes, geolayer_to_csv] +
        # driver.esri.shapefile
        [shapefile_to_geolayer, geolayer_to_shapefile] +
        # driver.geojson
        [feature_attributes_to_properties, geoformat_feature_to_geojson_feature, geolayer_to_geojson,
     json_object_to_feature_generator, from_geojson_get_features_list, geojson_to_geolayer] +
        # driver.prostgresql
        [fields_metadata_to_create_table_fields_structure, geometry_ref_to_geometry_field_structure,
     format_to_posgresq_values, geolayer_to_postgres] +
        # explore_data.duplicate
        [get_feature_hash, get_feature_geometry_hash, get_feature_attributes_hash, get_duplicate_features] +
        # explore_data.print_data
        [print_features_data_table, print_metadata_field_table] +
        # explore_data.random_geometry
        [random_point, random_segment, random_bbox] +
        # geoprocessing.connectors.operations
        [coordinates_to_point, coordinates_to_segment, coordinates_to_bbox, segment_to_bbox] +
        # geoprocessing.connectors.predicates
        [point_intersects_point, point_intersects_segment, point_intersects_bbox, segment_intersects_segment,
     segment_intersects_segment, segment_intersects_bbox, bbox_intersects_bbox, ccw_or_cw_segments, point_position_segment] +
        # geoprocessing.generalization
        [ramer_douglas_peucker, visvalingam_whyatt] +
        # geoprocessing.geoparameters.bbox
        [bbox_expand, bbox_union, point_bbox_position] +
        # geoprocessing.geoparameters.boundaries
        [ccw_or_cw_boundary] +
        # geoprocessing.geoparameters.lines
        [get_slope_between_two_points, get_intercept_from_point_and_slope, get_slope_from_point_and_intercept,
     line_parameters, perpendicular_line_parameters_at_point, point_at_distance_with_line_parameters,
     crossing_point_from_lines_parameters] +
        # geoprocession.matrix.adjacency
        [create_adjacency_matrix, get_area_intersecting_neighbors_i_feat, get_neighbor_i_feat] +
        # geoprocessing.measure.area
        [shoelace_formula, triangle_area] +
        # geoprocessing.measure.distance
        [euclidean_distance, manhattan_distance, euclidean_distance_point_vs_segment, point_vs_segment_distance] +
        # geoprocessing.measure.length
        [segment_length] +
        # geoprocessing.area
        [geometry_area] +
        # geoprocessing.length
        [geometry_length] +
        # geoprocessing.line_merge
        [line_merge] +
        # geoprocessing.merge_geometries
        [merge_geometries] +
        # geoprocessing.point_on_linestring
        [point_at_a_distance_on_segment, points_on_linestring_distance] +
        # geoprocessing.simplify
        [simplify] +
        # geoprocessing.split
        [segment_split_by_point, linestring_split_by_point] +
        # geoprocessing.union
        [union_by_split] +
        # index.attributes.hash
        [create_attribute_index] +
        # index.geometry.grid
        [bbox_to_g_id, point_to_g_id, g_id_to_bbox, g_id_to_point, g_id_neighbor_in_grid_index, create_grid_index,
     grid_index_to_geolayer] +
        # manipulation.feature_manipulation
        [rename_field_in_feature, drop_field_in_feature, drop_field_that_not_exists_in_feature] +
        # manipulation.geolayer_manipulation
        [add_attributes_index, check_attributes_index, check_if_field_exists, delete_attributes_index, delete_feature, drop_field, rename_field, create_field, split_geolayer_by_geometry_type] +
        # manipulation.metadata_manipulation
        [add_attributes_index_in_metadata, check_attributes_index_in_metadata, check_if_field_exists_in_metadata, delete_attributes_index_in_metadata, drop_field_in_metadata, drop_field_that_not_exists_in_metadata,  reorder_metadata_field_index_after_field_drop,
     rename_field_in_metadata, create_field_in_metadata] +
        # obj.geometry
        [len_coordinates, len_coordinates_in_geometry] +
        # processing.data.join.join
        [join, join_left, join_right, join_full] +
        # processing.data.join.merge_objects
        [merge_metadata, merge_feature] +
        # processing.data.clauses
        [clause_where, clause_group_by, clause_order_by] +
        # processing.data.fields_statistics
        [field_statistics] +
        # processing.data.union
        [union_metadata, union_geolayer]
)



from geoformat.conversion.geometry_conversion import (
    multi_geometry_to_single_geometry,
    geometry_to_geometry_collection
)
from geoformat.conversion.coordinates_conversion import format_coordinates
from geoformat.geoprocessing.connectors.operations import coordinates_to_bbox
from geoformat.geoprocessing.geoparameters.bbox import bbox_union


def line_merge(geometry, directed=False, bbox=False):
    """
    Returns a (set of) LineString(s) formed by sewing together the constituent line work of a MULTILINESTRING.

    :param geometry: input geomeetry that you want to merge. Obviously this function operate if you give Linestring or
    MultiLinestring in input, if not nothing append.
    :param directed: True if you won't reverse coordinates of merged geometry. False if you allowed reverse coordinates
    to realise a merge.
    :param bbox: True if you want add bbox to output geometry, False if not.

    """

    def merge_loop_over_single_linestring(multilinestring, directed):
        """
        This function define :
            * start_dict store each start point of a line part -> at each point you store the geom position in
            coordinates list.
            * end_dict store each end point point of a line part -> at each point you store the geom position in
            coordinates list.

            * hub_point with this two dictionaries we create hub_point that store for each unique point the counting
            number of times this point appears in the multilinestring.

            * we loop on each point in hub point and we try to know, if this point appears in two line, its position
            (start / end) for each of the lines to which it belongs. At this time you have 3 possibilities :
                # 2 points in start_dict
                # 2 points in end_dict
                # 1 point in start_dict and 1 point in end_dict

                * if 2 points in start_dict or 2 points in end_dict then we will reverse a linestring in input multi-
                linestring. If you choose directed is True then no reverse will be done (no merge either), if not the
                reverse is done and the merge is possible between two linestring.

                * we initialize some variables :
                    [before loop}
                    - final_coordinates (list) [before loop] : store each part to create a new geometry in output.
                    - computed_idx_geom (set) [before loop] : for each point on which we loop we store the under
                     geometry index.
                    - reloop (bool) [before loop] : if condition above (pass_point is True) then reloop is True.

                    [reinitialize at each loop step's]
                    - for_point_computed_idx_geom : store for each of two points the under geometry index. If we have
                    just one index inside it means we've processed a line that starts and ends at the same place
                    (a loop). In this case we don't change the geometry.
                    - result_part_coordinates (list) : store part new geometry.
                    - pass_point (bool) : if the geometry to which the point belongs has already been modified,
                     then it is not modified and pass_point is True.

                for each of 3 possibilities we create a new under geometry (result_part_coordinates) that we store step
                by step in final_ordinates if pass_point is False and for_point_computed_idx_geom.

            * Then we recreate a new geometry with data in final_coordinates and we add non concerned, by merging, geometries (geometries
            that aren't in computed_idx_geom).

            * We launch format coordinates to delete duplicate coordinates if exists
        """

        start_dict = {}
        end_dict = {}
        hub_point = {}

        single_geom_list = list(multi_geometry_to_single_geometry(multilinestring, bbox=False))

        for i_part_geom, single_geom in enumerate(single_geom_list):
            if single_geom['coordinates']:
                # begin
                start_point = tuple(single_geom['coordinates'][0])
                try:
                    start_dict[start_point].update({i_part_geom})
                except KeyError:
                    start_dict[start_point] = {i_part_geom}

                # end
                end_point = tuple(single_geom['coordinates'][-1])
                try:
                    end_dict[end_point].update({i_part_geom})
                except KeyError:
                    end_dict[end_point] = {i_part_geom}

                # count points occurrence
                # if count > 2 then we cannot merge geometries
                # count for start point
                hub_point = {point: len(start_dict[point]) for point in start_dict}
                # count for end point
                for end_point in end_dict:
                    if end_point in hub_point:
                        hub_point[end_point] = hub_point[end_point] + len(end_dict[end_point])
                    else:
                        hub_point[end_point] = len(end_dict[end_point])

        if hub_point:
            # loop over hub_point if at least one point exists more than one time we can merge
            # else we return input geometry
            to_merge = False
            for point, point_value in hub_point.items():
                if point_value > 1:
                    to_merge = True
                    break

            if to_merge is True:
                computed_idx_geom = set([])
                final_coordinates = []
                reloop = False

                for i_point, point in enumerate(hub_point):
                    # here three configurations are possible
                    # 2 points in start_dict
                    # 2 points in end_dict
                    # 1 point in start_dict and 1 point in end_dict
                    # As it possible to have a point in both dict (third configuration) and a point is always located
                    # between end of first geom and begin of second geom we begin with end_dict then start_dict

                    if hub_point[point] == 2:
                        if directed is True: # if directed is True and more than 1 point in end_dict or start_dict we
                                             # continue to next point in hub_point
                            if len(end_dict.get(point, [])) > 1 or len(start_dict.get(point, [])) > 1:
                                continue

                        result_part_coordinates = []
                        pass_point = False
                        for_point_computed_idx_geom = set([])
                        if point in end_dict:
                            if end_dict[point] & computed_idx_geom:
                                reloop = True
                                pass_point = True
                            else:
                                for i_idx, i_part_geom in enumerate(end_dict[point]):
                                    for_point_computed_idx_geom.update({i_part_geom})
                                    geom = multilinestring['coordinates'][i_part_geom]
                                    if i_idx == 0:
                                        result_part_coordinates += geom
                                    else:
                                        # reverse geom
                                        geom.reverse()
                                        # add to result
                                        result_part_coordinates = result_part_coordinates + geom[1:]

                        if point in start_dict:
                            if start_dict[point] & computed_idx_geom:
                                reloop = True
                                pass_point = True
                            else:
                                for i_idx, i_part_geom in enumerate(start_dict[point]):
                                    # add i_part_geom to computed_idx_geom
                                    for_point_computed_idx_geom.update({i_part_geom})
                                    geom = multilinestring['coordinates'][i_part_geom]
                                    if i_idx == 0:
                                        result_part_coordinates += geom[1:]
                                    else:
                                        # reverse geom
                                        geom.reverse()
                                        # add to result
                                        result_part_coordinates = geom + result_part_coordinates

                        # add result_part_coordinates to final_coordinates
                        if result_part_coordinates and not pass_point:
                            if len(for_point_computed_idx_geom) == 2:
                                final_coordinates.append(result_part_coordinates)
                            else:
                                i_part_geom = list(for_point_computed_idx_geom)[0]
                                geom = multilinestring['coordinates'][i_part_geom]
                                final_coordinates.append(geom)

                        # if this point is not passing, we add geom part in computed_idx_geom
                        if not pass_point:
                            computed_idx_geom.update(for_point_computed_idx_geom)

                # add non-computed idx geom
                for i_part_geom, single_geom in enumerate(single_geom_list):
                    if not {i_part_geom} & computed_idx_geom:
                        final_coordinates.append(single_geom['coordinates'])

                if len(final_coordinates) == 1:
                    final_geometry = {'type': 'LineString', 'coordinates': final_coordinates[0]}
                    # we do not have to reloop
                    reloop = False
                else:
                    final_geometry = {'type': 'MultiLineString', 'coordinates': final_coordinates}

                if reloop:
                    final_geometry = merge_loop_over_single_linestring(final_geometry, directed)

            #we return input geometry
            else:
                final_geometry = multilinestring
        else:
            final_geometry = multilinestring

        return final_geometry

    # init bbox for geometry collection
    bbox_geometry_collection = None

    # force input geometry to a geometry collection
    geometry_collection = geometry_to_geometry_collection(geometry, bbox=False)
    # loop on each geometry
    new_geometry_collection = {'type': 'GeometryCollection', 'geometries': list(geometry_collection['geometries'])}
    for i_geom, geometry_in_collection in enumerate(geometry_collection['geometries']):
        # if there is linestring in geometry_in_collection type, we launch a line_merge
        if "MULTILINESTRING" in geometry_in_collection['type'].upper():
            merged_geometry = merge_loop_over_single_linestring(geometry_in_collection, directed)
        else:
            merged_geometry = geometry_in_collection

        # format coordinates delete duplicates coordinates
        merged_geometry['coordinates'] = format_coordinates(coordinates_list_tuple=merged_geometry['coordinates'],
                                                            delete_duplicate_following_coordinates=True)
        # add bbox geometry if required
        if bbox and merged_geometry['coordinates']:
            if 'bbox' not in merged_geometry:
                merged_geometry['bbox'] = coordinates_to_bbox(merged_geometry['coordinates'])
            if i_geom == 0:
                bbox_geometry_collection = merged_geometry['bbox']
            else:
                bbox_geometry_collection = bbox_union(bbox_geometry_collection, merged_geometry['bbox'])

        new_geometry_collection['geometries'][i_geom] = merged_geometry

    # add bbox to a geometry collection
    if bbox and bbox_geometry_collection:
        new_geometry_collection['bbox'] = bbox_geometry_collection

    # if new_geometry_collection contain more than one geometry, we return a geometry collection
    if len(new_geometry_collection['geometries']) > 1:
        output_geometry_collection = new_geometry_collection
    # else we return a simple geometry or multi geometry
    else:
        # if geometry is not empty
        if new_geometry_collection['geometries']:
            output_geometry_collection = new_geometry_collection['geometries'][0]
        else:
            output_geometry_collection = new_geometry_collection

    return output_geometry_collection



import copy

from geoformat.conf.format_data import value_to_iterable_value
from geoformat.conversion.feature_conversion import (
    feature_filter
)
from geoformat.conversion.geometry_conversion import (
    multi_geometry_to_single_geometry,
    geometry_type_to_2d_geometry_type,
    geometry_to_2d_geometry,
    reproject_geometry
)

from geoformat.conversion.feature_conversion import feature_serialize, feature_deserialize

from geoformat.manipulation.metadata_manipulation import drop_field_that_not_exists_in_metadata


def multi_geometry_to_single_geometry_geolayer(geolayer):
    """
    Transform multi geometry geolayer (MultiPolygon, MultiLineString, MultiPoint) to single geometry geolayer (
    """
    # creation de l'output en copiant les metadata de l'input
    geolayer_out = copy.deepcopy(geolayer)
    bbox_extent = False
    del geolayer_out['features']
    if 'geometry_ref' in geolayer_out['metadata']:
        geolayer_out['metadata']['geometry_ref']['type'] = set()
        if 'extent' in geolayer_out['metadata']['geometry_ref']:
            bbox_extent = True

    # boucle et transformation des géométries multi part en single part
    new_i_feat = 0
    geolayer_out['features'] = {}
    for i_feat, feature in geolayer['features'].items():

        # if feature is serialized
        if 'feature_serialize' in geolayer['metadata']:
            if geolayer['metadata']['feature_serialize']:
                feature = feature_deserialize(serialized_feature=feature, bbox=False)

        geometry = feature['geometry']
        if geometry:
            for new_geometry in multi_geometry_to_single_geometry(geometry, bbox=bbox_extent):
                new_feature = {}
                if "attributes" in feature:
                    new_feature["attributes"] = feature['attributes']  # we add attributes first for better read

                new_feature["geometry"] = new_geometry


                # update geolayer metadata geometry type
                geolayer_out['metadata']['geometry_ref']['type'].update(set([new_geometry['type']]))

                # if feature is serialized
                if 'feature_serialize' in geolayer['metadata']:
                    if geolayer['metadata']['feature_serialize']:
                        new_feature = feature_serialize(feature=new_feature)

                geolayer_out['features'][new_i_feat] = new_feature
                new_i_feat += 1

    return geolayer_out


def geolayer_to_2d_geolayer(input_geolayer):
    """
    Transform geolayer with x dimension to geolayer who contain geometries with 2d dimension only.

    :param input_geolayer: geolayer to transform
    :return: 2d geometry geolayer
    """
    new_geolayer = {'metadata': copy.deepcopy(input_geolayer['metadata']), 'features': {}}

    # if geometry in geolayer
    if input_geolayer['metadata'].get('geometry_ref'):
        input_geometry_type = input_geolayer['metadata']['geometry_ref']['type']
        if isinstance(input_geometry_type, (list, tuple, set)):
            new_geometry_type = set()
            for geom_type in input_geometry_type:
                new_geometry_type.update([geometry_type_to_2d_geometry_type(geom_type)])
        else:
            new_geometry_type = geometry_type_to_2d_geometry_type(input_geometry_type)
        new_geolayer['metadata']['geometry_ref']['type'] = new_geometry_type

        if 'extent' in new_geolayer['metadata']['geometry_ref']:
            bbox_extent = True
        else:
            bbox_extent = False

    for i_feat, input_feature in input_geolayer['features'].items():
        if 'feature_serialize' in input_geolayer['metadata']:
            if input_geolayer['metadata']['feature_serialize'] == True:
                input_feature = eval(input_feature)

        output_feature = copy.deepcopy(input_feature)

        if 'geometry' in input_feature:
            input_geometry = input_feature['geometry']
            new_geometry = geometry_to_2d_geometry(input_geometry, bbox=bbox_extent)
            output_feature['geometry'] = new_geometry

        if 'feature_serialize' in input_geolayer['metadata']:
            if input_geolayer['metadata']['feature_serialize'] == True:
                output_feature = str(output_feature)

        new_geolayer['features'][i_feat] = output_feature

    return new_geolayer


def create_geolayer_from_i_feat_list(
        geolayer,
        i_feat_list,
        output_geolayer_name=None,
        field_name_filter=None,
        geometry_type_filter=None,
        bbox_filter=None,
        serialize=False,
        reset_i_feat=True
):
    """
    Create a new layer with i_feat_list from an input layer
    """
    i_feat_list = value_to_iterable_value(value=i_feat_list, output_iterable_type=list)

    new_geolayer_metadata = dict(geolayer['metadata'])
    if 'index' in new_geolayer_metadata:
        del new_geolayer_metadata['index']

     # rename geolayer
    if output_geolayer_name:
        new_geolayer_metadata['name'] = output_geolayer_name

    if field_name_filter:
        new_geolayer_metadata = drop_field_that_not_exists_in_metadata(
            metadata=new_geolayer_metadata,
            not_deleting_field_name_or_field_name_list=field_name_filter)

    if 'index' in new_geolayer_metadata:
        del new_geolayer_metadata['index']


    if geometry_type_filter:
        # erase geometry type in geometry ref
        if 'geometry_ref' in new_geolayer_metadata:
            new_geolayer_metadata['geometry_ref']['type'] = set()

    new_geolayer = {
        'metadata': new_geolayer_metadata
    }

    if serialize is True:
        new_geolayer['metadata']['feature_serialize'] = True

    new_geolayer['features'] = {}
    for i, i_feat in enumerate(i_feat_list):

        # if we have to reset i_feat
        if reset_i_feat:
            new_i_feat = i
        else:
            new_i_feat = i_feat

        # add feature to geolayer
        feature = geolayer['features'][i_feat]
        new_feature = feature_filter(
            feature=feature,
            serialize=serialize,
            field_name_filter=field_name_filter,
            geometry_type_filter=geometry_type_filter,
            bbox_filter=bbox_filter,
            bbox=False
        )
        if geometry_type_filter:
            # update geolayer geometry metadata
            if 'geometry' in new_feature:
                new_geolayer['metadata']['geometry_ref']['type'].update([new_feature['geometry']['type']])

        new_geolayer['features'][new_i_feat] = new_feature

    return new_geolayer


def reproject_geolayer(geolayer, out_crs, in_crs=None, precision=None):
    """
    Reproject geolayer from crs to an other

    :param geolayer: input geolayer.
    :param out_crs: new crs.
    :param in_crs:  input geolayer crs (if not present in geolayer metadata).
    :return:
    """

    geolayer = copy.deepcopy(geolayer)
    if not in_crs:
        in_crs = geolayer['metadata']['geometry_ref']['crs']

    # change metadata
    geolayer['metadata']['geometry_ref']['crs'] = out_crs


    # reproject geometry
    for i_feat in geolayer['features']:
        feature = geolayer['features'][i_feat]

        # if geometry in feature
        if 'geometry' in feature.keys():
            feature_geometry = feature['geometry']
            new_geometry = reproject_geometry(
                geometry=feature_geometry,
                in_crs=in_crs,
                out_crs=out_crs,
                precision=precision
            )

            # assign new geometry
            feature['geometry'] = new_geometry


    return geolayer



from geoformat.geoprocessing.measure.mesure_distance import (
    euclidean_distance,
    manhattan_distance
)

def segment_length(segment, distance_type='EUCLIDEAN'):
    """
    Calculate length of a given segment
    """
    (point_a, point_b) = segment
    if distance_type.upper() == 'EUCLIDEAN':
        return euclidean_distance(point_a, point_b)
    elif distance_type.upper() == 'MANHATTAN':
        return manhattan_distance(point_a, point_b)
    else:
        print('type of distance : {distance_type} does not exists'.format(distance_type=distance_type))



from geoformat.conversion.bbox_conversion import bbox_to_polygon_coordinates
from geoformat.conversion.feature_conversion import feature_serialize
from geoformat.geoprocessing.connectors.operations import coordinates_to_bbox
from geoformat.geoprocessing.geoparameters.bbox import point_bbox_position


########################################################################################################################
#
# GRID INDEX
#
########################################################################################################################


def bbox_to_g_id(bbox, mesh_size, x_grid_origin=0., y_grid_origin=0., grid_precision=None):
    """
    This function allow to find, with bbox coordinates, grid index identifier (g_id) to witch it begins.

    :param bbox: [list/tuple] bbox.
    :param mesh_size: [int] mesh size.
    :param x_grid_origin: x-coordinates of the original point from which the index is constructed.
    :param y_grid_origin: y-coordinates of the original point from which the index is constructed.
    :param grid_precision: precision of grid is the number of digits after the decimal point that we want keep.
    :return: g_id: generator of grid id (g_id) that bbox intersects.
    """

    if grid_precision:
        mesh_size = round(mesh_size, grid_precision)


    (x_min, y_min, x_max, y_max) = bbox
    # gives x_g_id, y_g_id for bbox extremity
    id_x_min = int((x_min - x_grid_origin) / mesh_size)
    if x_min - x_grid_origin < 0:
        id_x_min -= 1
    id_x_max = int((x_max - x_grid_origin) / mesh_size)
    if x_max - x_grid_origin < 0:
        id_x_max -= 1
    id_y_min = int((y_min - y_grid_origin) / mesh_size)
    if y_min - y_grid_origin < 0:
        id_y_min -= 1
    id_y_max = int((y_max - y_grid_origin) / mesh_size)
    if y_max - y_grid_origin < 0:
        id_y_max -= 1

    # recuperate g_id and bbox associate
    g_id_min = (id_x_min, id_y_min)
    g_id_max = (id_x_max, id_y_max)
    bbox_g_id_min = g_id_to_bbox(g_id_min, mesh_size, x_grid_origin=x_grid_origin, y_grid_origin=y_grid_origin, grid_precision=grid_precision)
    bbox_g_id_max = g_id_to_bbox(g_id_max, mesh_size, x_grid_origin=x_grid_origin, y_grid_origin=y_grid_origin, grid_precision=grid_precision)

    # test point position for x_min, y_min if point touch boundary we must change id_x_min and/or id_y_min
    point_position, point_direction = point_bbox_position((x_min, y_min), bbox_g_id_min)
    if point_position.upper() == 'BOUNDARY':  # 'Boundary':
        if point_direction.upper() == 'S':
            id_y_min += - 1
        if point_direction.upper() == 'W':
            id_x_min += - 1
        if point_direction.upper() == "NW":
            id_x_min += - 1
        if point_direction.upper() == "SW":
            id_x_min += - 1
            id_y_min += - 1
        if point_direction.upper() == "SE":
            id_y_min += - 1

    # test point position for x_max, y_max if point touch boundary we must change id_x_max and/or id_y_max
    point_position, point_direction = point_bbox_position((x_max, y_max), bbox_g_id_max)

    if point_position == 'Boundary':
        if point_direction == 'N':
            id_y_max += 1
        if point_direction == 'E':
            id_x_max += 1
        if point_direction == "NE":
            id_x_max += 1
            id_y_max += 1
        if point_direction == "NW":
            id_y_max += 1
        if point_direction == "SE":
            id_x_max += 1

    # double loop sendind all g_id intersecting bbox
    for x_step in range(id_x_min, id_x_max + 1):
        for y_step in range(id_y_min, id_y_max + 1):
            yield (x_step, y_step)


def point_to_g_id(point, mesh_size, x_grid_origin=0., y_grid_origin=0., grid_precision=None):
    """
    This function return grid id (g_id) for a given point.

    :param point: coordinates of point.
    :param mesh_size: size of mesh.
    :param x_grid_origin: coordinate in x of grid origin point.
    :param y_grid_origin: coordinate in y of grid origin point.
    :param grid_precision: precision of grid is the number of digits after the decimal point that we want keep.
    :return: a generator of g_id that intersects input point.
    """
    bbox = point + point

    for g_id in bbox_to_g_id(bbox, mesh_size, x_grid_origin, y_grid_origin, grid_precision=grid_precision):
        yield g_id


def g_id_to_bbox(g_id, mesh_size, x_grid_origin=0., y_grid_origin=0., grid_precision=None):
    """
    This fucntion return the bbox associate to a given g_id (grid id).
    If the grid origin is different to 0. you have to indicate the origin's coordinates (x_grid_origin or y_grid_origin)

    :param g_id: grid mesh identifier.
    :param mesh_size: size of mesh.
    :param x_grid_origin: coordinate in x of grid origin point.
    :param y_grid_origin: coordinate in y of grid origin point.
    :param grid_precision: precision of grid is the number of digits after the decimal point that we want keep.
    :return: bbox of g_id.
    """
    if grid_precision:
        mesh_size = round(mesh_size, grid_precision)
        x_grid_origin = round(x_grid_origin, grid_precision)
        y_grid_origin = round(y_grid_origin, grid_precision)

    (x_id, y_id) = g_id
    x_id, y_id = float(x_id), float(y_id)
    bbox_x_min = x_id * mesh_size
    bbox_x_max = (x_id + 1) * mesh_size
    bbox_y_min = y_id * mesh_size
    bbox_y_max = (y_id + 1) * mesh_size

    if x_grid_origin != 0. or y_grid_origin != 0.:
        bbox_x_min = x_grid_origin + bbox_x_min
        bbox_x_max = x_grid_origin + bbox_x_max
        bbox_y_min = y_grid_origin + bbox_y_min
        bbox_y_max = y_grid_origin + bbox_y_max

    if grid_precision:
        return round(bbox_x_min, grid_precision), round(bbox_y_min, grid_precision), round(bbox_x_max, grid_precision), round(bbox_y_max, grid_precision)
    else:
        return bbox_x_min, bbox_y_min, bbox_x_max, bbox_y_max


def g_id_to_point(g_id, mesh_size, position='center', x_grid_origin=0., y_grid_origin=0., grid_precision=None):
    """
    This function return a coordinates point to given g_id (g_id). Obviously since it is a grid, the position of the
    point can be specified between this values ('center', 'NE', 'NW', 'SE',  'SW').

    If the grid origin is different to 0. you have to indicate the origin's coordinates (x_grid_origin or y_grid_origin)

    :param g_id: g_id you want to get the point of.
    :param mesh_size: size of mesh.
    :param position: position of point in mesh (allowed values are ('center', 'NE', 'NW', 'SE',  'SW')).
    :param x_grid_origin: x coordinates of grid origin.
    :param y_grid_origin: y coordinates of grid origin.
    :param grid_precision: precision of grid is the number of digits after the decimal point that we want keep.
    :return: point coordinates.
    """
    x_min, y_min, x_max, y_max = g_id_to_bbox(g_id, mesh_size)
    position_upper = position.upper()

    if position_upper == 'CENTER':
        x, y = (x_min + x_max) / 2., (y_min + y_max) / 2.
    elif position_upper == 'NE':
        x, y = x_max, y_max
    elif position_upper == 'NW':
        x, y = x_min, y_max
    elif position_upper == 'SE':
        x, y = x_max, y_min
    elif position_upper == 'SW':
        x, y = x_min, y_min
    else:
        raise Exception('ERROR position must be a value between (CENTER, NE, NW, SE or SW)')

    if x_grid_origin != 0. or y_grid_origin != 0.:
        x += x_grid_origin
        y += y_grid_origin

    if grid_precision:
        x, y = round(x, grid_precision), round(y, grid_precision)

    return x, y


def g_id_neighbor_in_grid_index(g_id, grid_index, nb_mesh=1, g_id_include=True):
    """
    Return g_id neighbour if exists in given grid index at a distance of n mesh.

    :param g_id: mesh grid id.
    :param grid_index: grid index.
    :param nb_mesh: distance to find neighbour in mesh unit.
    :param g_id_include: if True output will return g_id value (input parameter).
    :return: generator of g_id neighbour.
    """
    (x_g_id, y_g_id) = g_id

    x_min = x_g_id - nb_mesh
    x_max = x_g_id + nb_mesh + 1
    y_min = y_g_id - nb_mesh
    y_max = y_g_id + nb_mesh + 1
    for x_g_id in range(x_min, x_max):
        for y_g_id in range(y_min, y_max):
            neighbor_g_id = x_g_id, y_g_id
            if neighbor_g_id in grid_index['index']:
                # if neighbor g_id is same that g_id and g_id_include is False we pass
                if g_id_include is False and neighbor_g_id == g_id:
                    pass
                else:
                    yield neighbor_g_id


def create_grid_index(geolayer, mesh_size=None, x_grid_origin=0, y_grid_origin=0, grid_precision=None):
    """
    Create a grid index for a geolayer at a mesh size (deduce automatically or given in input).
    You have also possibility to determine the origin point of the frame on which the grid is constructed (by default
    0, 0).

    This function return a dictionary with this structure : {
        'metadata : {
            'type' :  type of index
            'mesh_size': size of mesh (like in parameters or computed by this function if uninformed)
            'x_grid_origin':  x-coordinates of the original point from which the index is constructed.
                              By default, this value is 0.
            'y_grid_origin':  y-coordinates of the original point from which the index is constructed.
                              By default, this value is 0.
            'grid_precision': precision of grid is the number of digits after the decimal point that we want keep.
                              By default, this value is None.
        'index' : {
            (coordinates of mesh): [list with geolayer i_feat where bbox intersects mesh]
        }

    :param geolayer: geolayer that we want to index
    :param mesh_size: size of mesh
    :param x_grid_origin: x-coordinates of the original point from which the index is constructed.
    :param y_grid_origin: y-coordinates of the original point from which the index is constructed.
    :param grid_precision: precision of grid is the number of digits after the decimal point that we want keep.
     By default this value is None.
    :return: grid index dict (structure describe above)
    """

    # first define mesh size if not yet define
    if not mesh_size:
        # we deduce mean hight and width for all features
        size_x, size_y = 0, 0
        first_point = False
        for i, i_feat in enumerate(geolayer['features']):
            feature = geolayer['features'][i_feat]

            # if feature is serialized
            if 'feature_serialize' in geolayer['metadata']:
                if geolayer['metadata']['feature_serialize']:
                    feature = eval(feature)
            # get bbox
            try:
                bbox = feature['geometry']['bbox']
            except KeyError:
                bbox = coordinates_to_bbox(feature['geometry']['coordinates'])
                feature['geometry']['bbox'] = bbox

            # if geometry type is point there is no dimension in same bbox (no lenght / no width) we have to compare
            #  with others points bbox
            if feature['geometry']['type'] == 'Point':
                if not first_point:
                    old_point_bbox = bbox
                    first_point = True
                else:
                    size_x += abs(bbox[2] - old_point_bbox[2])
                    size_y += abs(bbox[3] - old_point_bbox[3])
                    old_point_bbox = bbox

            else:
                size_x += bbox[2] - bbox[0]
                size_y += bbox[3] - bbox[1]

        mesh_size = max(size_x / len(geolayer['features']), size_y / len(geolayer['features']))

    index_dict = {}
    # add metadata
    index_dict['metadata'] = {'type': 'grid',
                              'mesh_size': mesh_size,
                              'x_grid_origin': x_grid_origin,
                              'y_grid_origin': y_grid_origin,
                              'grid_precision': grid_precision}
    index_dict['index'] = {}

    for i_feat in geolayer['features']:
        feature = geolayer['features'][i_feat]

        # if feature is serialized
        if 'feature_serialize' in geolayer['metadata']:
            if geolayer['metadata']['feature_serialize']:
                feature = eval(feature)

        geom = feature['geometry']

        if 'bbox' in geom.keys():
            bbox_geom = geom['bbox']
        else:
            coordinates = geom['coordinates']
            bbox_geom = coordinates_to_bbox(coordinates)

        # for each grid identifier (g_id) we add in grid_idx_dico entity identifier (i_feat)
        for g_id in bbox_to_g_id(bbox=bbox_geom,
                                 mesh_size=mesh_size,
                                 x_grid_origin=x_grid_origin,
                                 y_grid_origin=y_grid_origin,
                                 grid_precision=grid_precision):
            try:
                index_dict['index'][g_id].append(i_feat)
            except KeyError:
                index_dict['index'][g_id] = [i_feat]

    return index_dict


def grid_index_to_geolayer(grid_index, name='grid_index', crs=None, features_serialize=False):
    """
    Create geolayer associate to grid index.

    :param grid_index: reference grid index
    :param name: name of geolayer
    :param crs: coordinates reference system
    :param features_serialize: True feature will be serialized / False feature are not serializes
    :return:  geolayer associate to grid index
    """

    geolayer = {'features': {}, 'metadata': {'name': name, 'geometry_ref': {'type': 'Polygon'}, }}

    if crs:
        geolayer['metadata']['geometry_ref']['crs'] = crs

    # create field
    geolayer['metadata']['fields'] = {'g_id': {'type': 'String', 'width': 254, 'index': 0},
                                      'i_feat_list': {'type': 'IntegerList', 'index': 1}}
    # add feature
    for i_feat, g_id in enumerate(grid_index['index']):
        feature = {
            'attributes': {'g_id': g_id, 'i_feat_list': grid_index['index'][g_id]},
            'geometry': {
                'type': 'Polygon',
                'coordinates': bbox_to_polygon_coordinates(
                    g_id_to_bbox(
                        g_id,
                        mesh_size=grid_index['metadata']['mesh_size'],
                        x_grid_origin=grid_index['metadata']['x_grid_origin'],
                        y_grid_origin=grid_index['metadata']['y_grid_origin']
                    )
                )
            }
        }

        if features_serialize:
            feature = feature_serialize(feature)

        geolayer['features'][i_feat] = feature

    return geolayer


GEOMETRY_CODE_TO_GEOMETRY_TYPE = {
    -2147483648: "25Bit",
    -2147483648: "25DBit",
    8: "CircularString",
    2008: "CircularStringM",
    1008: "CircularStringZ",
    3008: "CircularStringZM",
    9: "CompoundCurve",
    2009: "CompoundCurveM",
    1009: "CompoundCurveZ",
    3009: "CompoundCurveZM",
    13: "Curve",
    2013: "CurveM",
    10: "CurvePolygon",
    2010: "CurvePolygonM",
    1010: "CurvePolygonZ",
    3010: "CurvePolygonZM",
    1013: "CurveZ",
    3013: "CurveZM",
    7: "GeometryCollection",
    -2147483641: "GeometryCollection25D",
    2007: "GeometryCollectionM",
    3007: "GeometryCollectionZM",
    2: "LineString",
    -2147483646: "LineString25D",
    2002: "LineStringM",
    3002: "LineStringZM",
    101: "LinearRing",
    11: "MultiCurve",
    2011: "MultiCurveM",
    1011: "MultiCurveZ",
    3011: "MultiCurveZM",
    5: "MultiLineString",
    -2147483643: "MultiLineString25D",
    2005: "MultiLineStringM",
    3005: "MultiLineStringZM",
    4: "MultiPoint",
    -2147483644: "MultiPoint25D",
    2004: "MultiPointM",
    3004: "MultiPointZM",
    6: "MultiPolygon",
    -2147483642: "MultiPolygon25D",
    2006: "MultiPolygonM",
    3006: "MultiPolygonZM",
    12: "MultiSurface",
    2012: "MultiSurfaceM",
    1012: "MultiSurfaceZ",
    3012: "MultiSurfaceZM",
    100: "None",
    1: "Point",
    -2147483647: "Point25D",
    2001: "PointM",
    3001: "PointZM",
    3: "Polygon",
    -2147483645: "Polygon25D",
    2003: "PolygonM",
    3003: "PolygonZM",
    15: "PolyhedralSurface",
    2015: "PolyhedralSurfaceM",
    1015: "PolyhedralSurfaceZ",
    3015: "PolyhedralSurfaceZM",
    14: "Surface",
    2014: "SurfaceM",
    1014: "SurfaceZ",
    3014: "SurfaceZM",
    16: "TIN",
    2016: "TINM",
    1016: "TINZ",
    3016: "TINZM",
    17: "Triangle",
    2017: "TriangleM",
    1017: "TriangleZ",
    3017: "TriangleZM",
    0: "Unknown",
}

GEOMETRY_TYPE = {
    "25Bit": -2147483648,
    "25DBit": -2147483648,
    "CircularString": 8,
    "CircularStringM": 2008,
    "CircularStringZ": 1008,
    "CircularStringZM": 3008,
    "CompoundCurve": 9,
    "CompoundCurveM": 2009,
    "CompoundCurveZ": 1009,
    "CompoundCurveZM": 3009,
    "Curve": 13,
    "CurveM": 2013,
    "CurvePolygon": 10,
    "CurvePolygonM": 2010,
    "CurvePolygonZ": 1010,
    "CurvePolygonZM": 3010,
    "CurveZ": 1013,
    "CurveZM": 3013,
    "GeometryCollection": 7,
    "GeometryCollection25D": -2147483641,
    "GeometryCollectionM": 2007,
    "GeometryCollectionZM": 3007,
    "LineString": 2,
    "LineString25D": -2147483646,
    "LineStringM": 2002,
    "LineStringZM": 3002,
    "LinearRing": 101,
    "MultiCurve": 11,
    "MultiCurveM": 2011,
    "MultiCurveZ": 1011,
    "MultiCurveZM": 3011,
    "MultiLineString": 5,
    "MultiLineString25D": -2147483643,
    "MultiLineStringM": 2005,
    "MultiLineStringZM": 3005,
    "MultiPoint": 4,
    "MultiPoint25D": -2147483644,
    "MultiPointM": 2004,
    "MultiPointZM": 3004,
    "MultiPolygon": 6,
    "MultiPolygon25D": -2147483642,
    "MultiPolygonM": 2006,
    "MultiPolygonZM": 3006,
    "MultiSurface": 12,
    "MultiSurfaceM": 2012,
    "MultiSurfaceZ": 1012,
    "MultiSurfaceZM": 3012,
    "NDR": 1,
    "None": 100,
    "Point": 1,
    "Point25D": -2147483647,
    "PointM": 2001,
    "PointZM": 3001,
    "Polygon": 3,
    "Polygon25D": -2147483645,
    "PolygonM": 2003,
    "PolygonZM": 3003,
    "PolyhedralSurface": 15,
    "PolyhedralSurfaceM": 2015,
    "PolyhedralSurfaceZ": 1015,
    "PolyhedralSurfaceZM": 3015,
    "Surface": 14,
    "SurfaceM": 2014,
    "SurfaceZ": 1014,
    "SurfaceZM": 3014,
    "TIN": 16,
    "TINM": 2016,
    "TINZ": 1016,
    "TINZM": 3016,
    "Triangle": 17,
    "TriangleM": 2017,
    "TriangleZ": 1017,
    "TriangleZM": 3017,
    "Unknown": 0,
    "XDR": 0,
}

GEOFORMAT_GEOMETRY_TYPE = {
    'LINESTRING',
    'MULTIPOINT',
    'GEOMETRYCOLLECTION',
    'POINT',
    'GEOMETRY',
    'POLYGON',
    'MULTIPOLYGON',
    'MULTILINESTRING'}



from geoformat.conversion.geometry_conversion import (
    geometry_to_geometry_collection,
    multi_geometry_to_single_geometry
)
from geoformat.geoprocessing.measure.mesure_area import shoelace_formula


def geometry_area(geometry):
    # initialize return
    area = 0

    # transform function to geometry collection
    geometry = geometry_to_geometry_collection(geometry)
    # if geometry is not empty
    if geometry:
        for geometry in geometry['geometries']:
            # if geometry is not a point or multipoint (they are no dimension then no length)
            if 'POLYGON' in geometry['type'].upper():
                # loop over part
                for geometry_part in multi_geometry_to_single_geometry(geometry, bbox=False):
                    coordinates_list = geometry_part['coordinates']
                    # if there is coordinates
                    if coordinates_list:
                        for i_ring, coordinates_ring in enumerate(coordinates_list):
                            ring_area = shoelace_formula(coordinates_ring)
                            if i_ring == 0:
                                area += ring_area
                            else:
                                area -= ring_area
        return area




def coordinates_to_point(coordinates, ):
    """
    With geometry coordinates this function return each point individually in iterator
    :param coordinates
    :yield point coordinate
    """

    if isinstance(coordinates[0], (int, float)):
        yield list(coordinates)
    else:
        for inner_coordinates in coordinates:
            for point in coordinates_to_point(inner_coordinates):
                yield point


def coordinates_to_segment(coordinates):
    """
    This function yield
    :param coordinates:
    :return:
    """
    if isinstance(coordinates[0][0], (float, int)):
        for i, point in enumerate(coordinates_to_point(coordinates)):
            if i == 0:
                pass
            else:
                yield [save_point, point]
            save_point = point
    else:
        for part_coordinates in coordinates:
            for segment in coordinates_to_segment(part_coordinates):
                yield segment


def coordinates_to_bbox(coordinates):
    """
    This function return boundaries box (bbox) from geometry coordinates.
    It's works with 2 to n-dimensional data.
    (x_min, y_min, n_min, x_max, y_max, n_max)

    """
    # if list is not empty
    if coordinates:
        # loop on each coordinates
        bbox = None
        for i_pt, point in enumerate(coordinates_to_point(coordinates)):
            nb_coord = len(point)
            # create bbox structure
            if bbox is None:
                bbox = point * nb_coord
            else:
                # add min and max value of bbox for other coordinates
                for i_coord, coord in enumerate(point):
                    if coord < bbox[i_coord]:
                        bbox[i_coord] = coord
                    if coord > bbox[i_coord + nb_coord]:
                        bbox[i_coord + nb_coord] = coord

        return tuple(bbox)
    else:
        return ()


def segment_to_bbox(segment):
    """
    Return bbox to given segment

    :param segment: segment of which we want to know the bbox
    :return: bbox of segment
    """
    x, y = zip(*segment)

    return min(x), min(y), max(x), max(y)
